# -*- coding: utf-8 -*-
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import unittest

from compare_locales.paths import ProjectConfig, File, ProjectFiles, Matcher
from compare_locales import mozpath


class TestMatcher(unittest.TestCase):

    def test_matcher(self):
        one = Matcher('foo/*')
        self.assertTrue(one.match('foo/baz'))
        self.assertFalse(one.match('foo/baz/qux'))
        other = Matcher('bar/*')
        self.assertTrue(other.match('bar/baz'))
        self.assertFalse(other.match('bar/baz/qux'))
        self.assertEqual(one.sub(other, 'foo/baz'), 'bar/baz')
        self.assertIsNone(one.sub(other, 'bar/baz'))
        one = Matcher('foo/**')
        self.assertTrue(one.match('foo/baz'))
        self.assertTrue(one.match('foo/baz/qux'))
        other = Matcher('bar/**')
        self.assertTrue(other.match('bar/baz'))
        self.assertTrue(other.match('bar/baz/qux'))
        self.assertEqual(one.sub(other, 'foo/baz'), 'bar/baz')
        self.assertEqual(one.sub(other, 'foo/baz/qux'), 'bar/baz/qux')
        one = Matcher('foo/*/one/**')
        self.assertTrue(one.match('foo/baz/one/qux'))
        self.assertFalse(one.match('foo/baz/bez/one/qux'))
        other = Matcher('bar/*/other/**')
        self.assertTrue(other.match('bar/baz/other/qux'))
        self.assertFalse(other.match('bar/baz/bez/other/qux'))
        self.assertEqual(one.sub(other, 'foo/baz/one/qux'),
                         'bar/baz/other/qux')
        self.assertEqual(one.sub(other, 'foo/baz/one/qux/zzz'),
                         'bar/baz/other/qux/zzz')
        self.assertIsNone(one.sub(other, 'foo/baz/bez/one/qux'))


class SetupMixin(object):
    def setUp(self):
        self.cfg = ProjectConfig()
        self.file = File(
            '/tmp/somedir/de/browser/one/two/file.ftl',
            'file.ftl',
            module='browser', locale='de')
        self.other_file = File(
            '/tmp/somedir/de/toolkit/two/one/file.ftl',
            'file.ftl',
            module='toolkit', locale='de')


class TestConfigLegacy(SetupMixin, unittest.TestCase):

    def test_filter_py_true(self):
        'Test filter.py just return bool(True)'
        def filter(mod, path, entity=None):
            return True
        self.cfg.set_filter_py(filter)
        with self.assertRaises(AssertionError):
            self.cfg.add_rules({})
        rv = self.cfg.filter(self.file)
        self.assertEqual(rv, 'error')
        rv = self.cfg.filter(self.file, entity='one_entity')
        self.assertEqual(rv, 'error')

    def test_filter_py_false(self):
        'Test filter.py just return bool(False)'
        def filter(mod, path, entity=None):
            return False
        self.cfg.set_filter_py(filter)
        with self.assertRaises(AssertionError):
            self.cfg.add_rules({})
        rv = self.cfg.filter(self.file)
        self.assertEqual(rv, 'ignore')
        rv = self.cfg.filter(self.file, entity='one_entity')
        self.assertEqual(rv, 'ignore')

    def test_filter_py_error(self):
        'Test filter.py just return str("error")'
        def filter(mod, path, entity=None):
            return 'error'
        self.cfg.set_filter_py(filter)
        with self.assertRaises(AssertionError):
            self.cfg.add_rules({})
        rv = self.cfg.filter(self.file)
        self.assertEqual(rv, 'error')
        rv = self.cfg.filter(self.file, entity='one_entity')
        self.assertEqual(rv, 'error')

    def test_filter_py_ignore(self):
        'Test filter.py just return str("ignore")'
        def filter(mod, path, entity=None):
            return 'ignore'
        self.cfg.set_filter_py(filter)
        with self.assertRaises(AssertionError):
            self.cfg.add_rules({})
        rv = self.cfg.filter(self.file)
        self.assertEqual(rv, 'ignore')
        rv = self.cfg.filter(self.file, entity='one_entity')
        self.assertEqual(rv, 'ignore')

    def test_filter_py_report(self):
        'Test filter.py just return str("report") and match to "warning"'
        def filter(mod, path, entity=None):
            return 'report'
        self.cfg.set_filter_py(filter)
        with self.assertRaises(AssertionError):
            self.cfg.add_rules({})
        rv = self.cfg.filter(self.file)
        self.assertEqual(rv, 'warning')
        rv = self.cfg.filter(self.file, entity='one_entity')
        self.assertEqual(rv, 'warning')

    def test_filter_py_module(self):
        'Test filter.py to return str("error") for browser or "ignore"'
        def filter(mod, path, entity=None):
            return 'error' if mod == 'browser' else 'ignore'
        self.cfg.set_filter_py(filter)
        with self.assertRaises(AssertionError):
            self.cfg.add_rules({})
        rv = self.cfg.filter(self.file)
        self.assertEqual(rv, 'error')
        rv = self.cfg.filter(self.file, entity='one_entity')
        self.assertEqual(rv, 'error')
        rv = self.cfg.filter(self.other_file)
        self.assertEqual(rv, 'ignore')
        rv = self.cfg.filter(self.other_file, entity='one_entity')
        self.assertEqual(rv, 'ignore')


class TestConfigRules(SetupMixin, unittest.TestCase):

    def test_filter_empty(self):
        'Test that an empty config works'
        self.cfg.add_paths({
            'l10n': '/tmp/somedir/{locale}/browser/**'
        })
        rv = self.cfg.filter(self.file)
        self.assertEqual(rv, 'error')
        rv = self.cfg.filter(self.file, entity='one_entity')
        self.assertEqual(rv, 'error')
        rv = self.cfg.filter(self.other_file)
        self.assertEqual(rv, 'ignore')
        rv = self.cfg.filter(self.other_file, entity='one_entity')
        self.assertEqual(rv, 'ignore')

    def test_single_file_rule(self):
        'Test a single rule for just a single file, no key'
        self.cfg.add_paths({
            'l10n': '/tmp/somedir/{locale}/browser/**'
        })
        self.cfg.add_rules({
            'path': '/tmp/somedir/{locale}/browser/one/two/file.ftl',
            'action': 'ignore'
        })
        rv = self.cfg.filter(self.file)
        self.assertEqual(rv, 'ignore')
        rv = self.cfg.filter(self.file, 'one_entity')
        self.assertEqual(rv, 'error')
        rv = self.cfg.filter(self.other_file)
        self.assertEqual(rv, 'ignore')
        rv = self.cfg.filter(self.other_file, 'one_entity')
        self.assertEqual(rv, 'ignore')

    def test_single_key_rule(self):
        'Test a single rule with file and key'
        self.cfg.add_paths({
            'l10n': '/tmp/somedir/{locale}/browser/**'
        })
        self.cfg.add_rules({
            'path': '/tmp/somedir/{locale}/browser/one/two/file.ftl',
            'key': 'one_entity',
            'action': 'ignore'
        })
        rv = self.cfg.filter(self.file)
        self.assertEqual(rv, 'error')
        rv = self.cfg.filter(self.file, 'one_entity')
        self.assertEqual(rv, 'ignore')
        rv = self.cfg.filter(self.other_file)
        self.assertEqual(rv, 'ignore')
        rv = self.cfg.filter(self.other_file, 'one_entity')
        self.assertEqual(rv, 'ignore')

    def test_single_non_matching_key_rule(self):
        'Test a single key rule with regex special chars that should not match'
        self.cfg.add_paths({
            'l10n': '/tmp/somedir/{locale}/**'
        })
        self.cfg.add_rules({
            'path': '/tmp/somedir/{locale}/browser/one/two/file.ftl',
            'key': '.ne_entit.',
            'action': 'ignore'
        })
        rv = self.cfg.filter(self.file, 'one_entity')
        self.assertEqual(rv, 'error')

    def test_single_matching_re_key_rule(self):
        'Test a single key with regular expression'
        self.cfg.add_paths({
            'l10n': '/tmp/somedir/{locale}/**'
        })
        self.cfg.add_rules({
            'path': '/tmp/somedir/{locale}/browser/one/two/file.ftl',
            'key': 're:.ne_entit.$',
            'action': 'ignore'
        })
        rv = self.cfg.filter(self.file, 'one_entity')
        self.assertEqual(rv, 'ignore')

    def test_double_file_rule(self):
        'Test path shortcut, one for each of our files'
        self.cfg.add_paths({
            'l10n': '/tmp/somedir/{locale}/**'
        })
        self.cfg.add_rules({
            'path': [
                '/tmp/somedir/{locale}/browser/one/two/file.ftl',
                '/tmp/somedir/{locale}/toolkit/two/one/file.ftl',
            ],
            'action': 'ignore'
        })
        rv = self.cfg.filter(self.file)
        self.assertEqual(rv, 'ignore')
        rv = self.cfg.filter(self.other_file)
        self.assertEqual(rv, 'ignore')

    def test_double_file_key_rule(self):
        'Test path and key shortcut, one key matching, one not'
        self.cfg.add_paths({
            'l10n': '/tmp/somedir/{locale}/**'
        })
        self.cfg.add_rules({
            'path': [
                '/tmp/somedir/{locale}/browser/one/two/file.ftl',
                '/tmp/somedir/{locale}/toolkit/two/one/file.ftl',
            ],
            'key': [
                'one_entity',
                'other_entity',
            ],
            'action': 'ignore'
        })
        rv = self.cfg.filter(self.file)
        self.assertEqual(rv, 'error')
        rv = self.cfg.filter(self.file, 'one_entity')
        self.assertEqual(rv, 'ignore')
        rv = self.cfg.filter(self.other_file)
        self.assertEqual(rv, 'error')
        rv = self.cfg.filter(self.other_file, 'one_entity')
        self.assertEqual(rv, 'ignore')

    def test_single_wildcard_rule(self):
        'Test single wildcard'
        self.cfg.add_paths({
            'l10n': '/tmp/somedir/{locale}/browser/**'
        })
        self.cfg.add_rules({
            'path': [
                '/tmp/somedir/{locale}/browser/one/*/*',
            ],
            'action': 'ignore'
        })
        rv = self.cfg.filter(self.file)
        self.assertEqual(rv, 'ignore')
        rv = self.cfg.filter(self.other_file)
        self.assertEqual(rv, 'ignore')

    def test_double_wildcard_rule(self):
        'Test double wildcard'
        self.cfg.add_paths({
            'l10n': '/tmp/somedir/{locale}/**'
        })
        self.cfg.add_rules({
            'path': [
                '/tmp/somedir/{locale}/**',
            ],
            'action': 'ignore'
        })
        rv = self.cfg.filter(self.file)
        self.assertEqual(rv, 'ignore')
        rv = self.cfg.filter(self.other_file)
        self.assertEqual(rv, 'ignore')


class MockProjectFiles(ProjectFiles):
    def __init__(self, mocks, locale, projects, mergebase=None):
        (super(MockProjectFiles, self)
            .__init__(locale, projects, mergebase=mergebase))
        self.mocks = mocks

    def _files(self, matcher):
        base = matcher.prefix
        for path in self.mocks.get(base, []):
            p = mozpath.join(base, path)
            if matcher.match(p):
                yield p


class TestProjectPaths(unittest.TestCase):
    def test_l10n_path(self):
        cfg = ProjectConfig()
        cfg.locales.append('de')
        cfg.add_paths({
            'l10n': '{l10n_base}/{locale}/*'
        })
        cfg.add_environment(l10n_base='/tmp')
        mocks = {
            '/tmp/de/': [
                'good.ftl',
                'not/subdir/bad.ftl'
            ],
            '/tmp/fr/': [
                'good.ftl',
                'not/subdir/bad.ftl'
            ],
        }
        files = MockProjectFiles(mocks, 'de', [cfg])
        self.assertListEqual(
            list(files), [('/tmp/de/good.ftl', None, None, set())])
        self.assertTupleEqual(
            files.match('/tmp/de/something.ftl'),
            ('/tmp/de/something.ftl', None, None, set()))
        self.assertIsNone(files.match('/tmp/fr/something.ftl'))
        files = MockProjectFiles(mocks, 'de', [cfg], mergebase='merging')
        self.assertListEqual(
            list(files),
            [('/tmp/de/good.ftl', None, 'merging/de/good.ftl', set())])
        self.assertTupleEqual(
            files.match('/tmp/de/something.ftl'),
            ('/tmp/de/something.ftl', None, 'merging/de/something.ftl', set()))
        # 'fr' is not in the locale list, should return no files
        files = MockProjectFiles(mocks, 'fr', [cfg])
        self.assertListEqual(list(files), [])

    def test_reference_path(self):
        cfg = ProjectConfig()
        cfg.locales.append('de')
        cfg.add_paths({
            'l10n': '{l10n_base}/{locale}/*',
            'reference': '/tmp/reference/*'
        })
        cfg.add_environment(l10n_base='/tmp/l10n')
        mocks = {
            '/tmp/l10n/de/': [
                'good.ftl',
                'not/subdir/bad.ftl'
            ],
            '/tmp/l10n/fr/': [
                'good.ftl',
                'not/subdir/bad.ftl'
            ],
            '/tmp/reference/': [
                'ref.ftl',
                'not/subdir/bad.ftl'
            ],
        }
        files = MockProjectFiles(mocks, 'de', [cfg])
        self.assertListEqual(
            list(files),
            [
                ('/tmp/l10n/de/good.ftl', '/tmp/reference/good.ftl', None,
                 set()),
                ('/tmp/l10n/de/ref.ftl', '/tmp/reference/ref.ftl', None,
                 set()),
            ])
        self.assertTupleEqual(
            files.match('/tmp/l10n/de/good.ftl'),
            ('/tmp/l10n/de/good.ftl', '/tmp/reference/good.ftl', None,
             set()),
            )
        self.assertTupleEqual(
            files.match('/tmp/reference/good.ftl'),
            ('/tmp/l10n/de/good.ftl', '/tmp/reference/good.ftl', None,
             set()),
            )
        self.assertIsNone(files.match('/tmp/l10n/de/subdir/bad.ftl'))
        self.assertIsNone(files.match('/tmp/reference/subdir/bad.ftl'))
        files = MockProjectFiles(mocks, 'de', [cfg], mergebase='merging')
        self.assertListEqual(
            list(files),
            [
                ('/tmp/l10n/de/good.ftl', '/tmp/reference/good.ftl',
                 'merging/de/good.ftl', set()),
                ('/tmp/l10n/de/ref.ftl', '/tmp/reference/ref.ftl',
                 'merging/de/ref.ftl', set()),
            ])
        self.assertTupleEqual(
            files.match('/tmp/l10n/de/good.ftl'),
            ('/tmp/l10n/de/good.ftl', '/tmp/reference/good.ftl',
             'merging/de/good.ftl', set()),
            )
        self.assertTupleEqual(
            files.match('/tmp/reference/good.ftl'),
            ('/tmp/l10n/de/good.ftl', '/tmp/reference/good.ftl',
             'merging/de/good.ftl', set()),
            )
        # 'fr' is not in the locale list, should return no files
        files = MockProjectFiles(mocks, 'fr', [cfg])
        self.assertListEqual(list(files), [])

    def test_partial_l10n(self):
        cfg = ProjectConfig()
        cfg.locales.extend(['de', 'fr'])
        cfg.add_paths({
            'l10n': '/tmp/{locale}/major/*'
        }, {
            'l10n': '/tmp/{locale}/minor/*',
            'locales': ['de']
        })
        mocks = {
            '/tmp/de/major/': [
                'good.ftl',
                'not/subdir/bad.ftl'
            ],
            '/tmp/de/minor/': [
                'good.ftl',
            ],
            '/tmp/fr/major/': [
                'good.ftl',
                'not/subdir/bad.ftl'
            ],
            '/tmp/fr/minor/': [
                'good.ftl',
            ],
        }
        files = MockProjectFiles(mocks, 'de', [cfg])
        self.assertListEqual(
            list(files),
            [
                ('/tmp/de/major/good.ftl', None, None, set()),
                ('/tmp/de/minor/good.ftl', None, None, set()),
            ])
        self.assertTupleEqual(
            files.match('/tmp/de/major/some.ftl'),
            ('/tmp/de/major/some.ftl', None, None, set()))
        self.assertIsNone(files.match('/tmp/de/other/some.ftl'))
        # 'fr' is not in the locale list of minor, should only return major
        files = MockProjectFiles(mocks, 'fr', [cfg])
        self.assertListEqual(
            list(files),
            [
                ('/tmp/fr/major/good.ftl', None, None, set()),
            ])
        self.assertIsNone(files.match('/tmp/fr/minor/some.ftl'))


class TestProjectConfig(unittest.TestCase):
    def test_expand_paths(self):
        pc = ProjectConfig()
        pc.add_environment(one="first_path")
        self.assertEqual(pc.expand('foo'), 'foo')
        self.assertEqual(pc.expand('foo{one}bar'), 'foofirst_pathbar')
        pc.add_environment(l10n_base='../tmp/localizations')
        self.assertEqual(
            pc.expand('{l}dir', {'l': '{l10n_base}/{locale}/'}),
            '../tmp/localizations/{locale}/dir')
        self.assertEqual(
            pc.expand('{l}dir', {
                'l': '{l10n_base}/{locale}/',
                'l10n_base': '../merge-base'
            }),
            '../merge-base/{locale}/dir')

    def test_children(self):
        pc = ProjectConfig()
        child = ProjectConfig()
        pc.add_child(child)
        self.assertListEqual([pc, child], list(pc.configs))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import, print_function, unicode_literals

import unittest

from taskgraph import morph
from taskgraph.graph import Graph
from taskgraph.taskgraph import TaskGraph
from taskgraph.task import Task

from mozunit import main


class MorphTestCase(unittest.TestCase):

    def make_taskgraph(self, tasks):
        label_to_taskid = {k: k + '-tid' for k in tasks}
        for label, task_id in label_to_taskid.iteritems():
            tasks[label].task_id = task_id
        graph = Graph(nodes=set(tasks), edges=set())
        taskgraph = TaskGraph(tasks, graph)
        return taskgraph, label_to_taskid


class TestIndexTask(MorphTestCase):

    def test_make_index_tasks(self):
        task_def = {
            'routes': [
                "index.gecko.v2.mozilla-central.latest.firefox-l10n.linux64-opt.es-MX",
                "index.gecko.v2.mozilla-central.latest.firefox-l10n.linux64-opt.fy-NL",
                "index.gecko.v2.mozilla-central.latest.firefox-l10n.linux64-opt.sk",
                "index.gecko.v2.mozilla-central.latest.firefox-l10n.linux64-opt.sl",
                "index.gecko.v2.mozilla-central.latest.firefox-l10n.linux64-opt.uk",
                "index.gecko.v2.mozilla-central.latest.firefox-l10n.linux64-opt.zh-CN",
                "index.gecko.v2.mozilla-central.pushdate."
                "2017.04.04.20170404100210.firefox-l10n.linux64-opt.es-MX",
                "index.gecko.v2.mozilla-central.pushdate."
                "2017.04.04.20170404100210.firefox-l10n.linux64-opt.fy-NL",
                "index.gecko.v2.mozilla-central.pushdate."
                "2017.04.04.20170404100210.firefox-l10n.linux64-opt.sk",
                "index.gecko.v2.mozilla-central.pushdate."
                "2017.04.04.20170404100210.firefox-l10n.linux64-opt.sl",
                "index.gecko.v2.mozilla-central.pushdate."
                "2017.04.04.20170404100210.firefox-l10n.linux64-opt.uk",
                "index.gecko.v2.mozilla-central.pushdate."
                "2017.04.04.20170404100210.firefox-l10n.linux64-opt.zh-CN",
                "index.gecko.v2.mozilla-central.revision."
                "b5d8b27a753725c1de41ffae2e338798f3b5cacd.firefox-l10n.linux64-opt.es-MX",
                "index.gecko.v2.mozilla-central.revision."
                "b5d8b27a753725c1de41ffae2e338798f3b5cacd.firefox-l10n.linux64-opt.fy-NL",
                "index.gecko.v2.mozilla-central.revision."
                "b5d8b27a753725c1de41ffae2e338798f3b5cacd.firefox-l10n.linux64-opt.sk",
                "index.gecko.v2.mozilla-central.revision."
                "b5d8b27a753725c1de41ffae2e338798f3b5cacd.firefox-l10n.linux64-opt.sl",
                "index.gecko.v2.mozilla-central.revision."
                "b5d8b27a753725c1de41ffae2e338798f3b5cacd.firefox-l10n.linux64-opt.uk",
                "index.gecko.v2.mozilla-central.revision."
                "b5d8b27a753725c1de41ffae2e338798f3b5cacd.firefox-l10n.linux64-opt.zh-CN"
            ],
            'deadline': 'soon',
            'metadata': {
                'description': 'desc',
                'owner': 'owner@foo.com',
                'source': 'https://source',
            },
        }
        task = Task(kind='test', label='a', attributes={}, task=task_def)
        docker_task = Task(kind='docker-image', label='build-docker-image-index-task',
                           attributes={}, task={})
        taskgraph, label_to_taskid = self.make_taskgraph({
            task.label: task,
            docker_task.label: docker_task,
        })

        index_task = morph.make_index_task(task, taskgraph, label_to_taskid)

        self.assertEqual(index_task.task['payload']['command'][0], 'insert-indexes.js')
        self.assertEqual(index_task.task['payload']['env']['TARGET_TASKID'], 'a-tid')

        # check the scope summary
        self.assertEqual(index_task.task['scopes'],
                         ['index:insert-task:gecko.v2.mozilla-central.*'])


class TestApplyJSONeTemplates(MorphTestCase):

    tasks = [
        {
            'kind': 'build',
            'label': 'a',
            'attributes': {},
            'task': {
                'extra': {
                    'treeherder': {
                        'group': 'tc',
                        'symbol': 'B'
                    }
                },
                'payload': {
                    'env': {
                        'FOO': 'BAR'
                    }
                },
                'tags': {
                    'kind': 'build'
                }
            }
        },
        {
            'kind': 'test',
            'label': 'b',
            'attributes': {},
            'task': {
                'extra': {
                    'treeherder': {
                        'group': 'tc',
                        'symbol': 't'
                    }
                },
                'payload': {
                    'env': {
                        'FOO': 'BAR'
                    }
                },
                'tags': {
                    'kind': 'test'
                }
            }
        },
    ]

    def test_template_artifact(self):
        tg, label_to_taskid = self.make_taskgraph({
            t['label']: Task(**t) for t in self.tasks[:]
        })

        try_task_config = {
            'templates': {
                'artifact': {'enabled': 1}
            },
        }
        fn = morph.apply_jsone_templates(try_task_config)
        morphed = fn(tg, label_to_taskid)[0]

        self.assertEqual(len(morphed.tasks), 2)

        for t in morphed.tasks.values():
            if t.kind == 'build':
                self.assertEqual(t.task['extra']['treeherder']['group'], 'tc')
                self.assertEqual(t.task['extra']['treeherder']['symbol'], 'Ba')
                self.assertEqual(t.task['payload']['env']['USE_ARTIFACT'], 1)
            else:
                self.assertEqual(t.task['extra']['treeherder']['group'], 'tc')
                self.assertEqual(t.task['extra']['treeherder']['symbol'], 't')
                self.assertNotIn('USE_ARTIFACT', t.task['payload']['env'])

    def test_template_env(self):
        tg, label_to_taskid = self.make_taskgraph({
            t['label']: Task(**t) for t in self.tasks[:]
        })

        try_task_config = {
            'templates': {
                'env': {
                    'ENABLED': 1,
                    'FOO': 'BAZ',
                }
            },
        }
        fn = morph.apply_jsone_templates(try_task_config)
        morphed = fn(tg, label_to_taskid)[0]

        self.assertEqual(len(morphed.tasks), 2)
        for t in morphed.tasks.values():
            self.assertEqual(len(t.task['payload']['env']), 2)
            self.assertEqual(t.task['payload']['env']['ENABLED'], 1)
            self.assertEqual(t.task['payload']['env']['FOO'], 'BAZ')

        try_task_config['templates']['env'] = {
            'ENABLED': 0,
        }
        fn = morph.apply_jsone_templates(try_task_config)
        morphed = fn(tg, label_to_taskid)[0]

        self.assertEqual(len(morphed.tasks), 2)
        for t in morphed.tasks.values():
            self.assertEqual(len(t.task['payload']['env']), 2)
            self.assertEqual(t.task['payload']['env']['ENABLED'], 0)
            self.assertEqual(t.task['payload']['env']['FOO'], 'BAZ')

    def test_template_rebuild(self):
        tg, label_to_taskid = self.make_taskgraph({
            t['label']: Task(**t) for t in self.tasks[:]
        })

        try_task_config = {
            'tasks': ['b'],
            'templates': {
                'rebuild': 4,
            },
        }
        fn = morph.apply_jsone_templates(try_task_config)
        tasks = fn(tg, label_to_taskid)[0].tasks.values()
        self.assertEqual(len(tasks), 2)

        for t in tasks:
            if t.label == 'a':
                self.assertNotIn('task_duplicates', t.attributes)
            elif t.label == 'b':
                self.assertIn('task_duplicates', t.attributes)
                self.assertEqual(t.attributes['task_duplicates'], 4)


if __name__ == '__main__':
    main()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""Pytest fixtures to help set up Firefox and a tests.zip
in test harness selftests.
"""

from __future__ import absolute_import

import os
import shutil
import sys

import mozfile
import mozinstall
import pytest
import requests
from mozbuild.base import MozbuildObject

here = os.path.abspath(os.path.dirname(__file__))
build = MozbuildObject.from_environment(cwd=here)


HARNESS_ROOT_NOT_FOUND = """
Could not find test harness root. Either a build or the 'GECKO_INSTALLER_URL'
environment variable is required.
""".lstrip()


def _get_test_harness(suite, install_dir):
    # Check if there is a local build
    harness_root = os.path.join(build.topobjdir, '_tests', install_dir)
    if os.path.isdir(harness_root):
        return harness_root

    # Check if it was previously set up by another test
    harness_root = os.path.join(os.environ['PYTHON_TEST_TMP'], 'tests', suite)
    if os.path.isdir(harness_root):
        return harness_root

    # Check if there is a test package to download
    if 'GECKO_INSTALLER_URL' in os.environ:
        base_url = os.environ['GECKO_INSTALLER_URL'].rsplit('/', 1)[0]
        test_packages = requests.get(base_url + '/target.test_packages.json').json()

        dest = os.path.join(os.environ['PYTHON_TEST_TMP'], 'tests')
        for name in test_packages[suite]:
            url = base_url + '/' + name
            bundle = os.path.join(os.environ['PYTHON_TEST_TMP'], name)

            r = requests.get(url, stream=True)
            with open(bundle, 'w+b') as fh:
                for chunk in r.iter_content(chunk_size=1024):
                    fh.write(chunk)

            mozfile.extract(bundle, dest)

        return os.path.join(dest, suite)

    # Couldn't find a harness root, let caller do error handling.
    return None


@pytest.fixture(scope='session')
def setup_test_harness(request):
    """Fixture for setting up a mozharness-based test harness like
    mochitest or reftest"""
    def inner(files_dir, *args, **kwargs):
        harness_root = _get_test_harness(*args, **kwargs)
        if harness_root:
            sys.path.insert(0, harness_root)

            # Link the test files to the test package so updates are automatically
            # picked up. Fallback to copy on Windows.
            if files_dir:
                test_root = os.path.join(harness_root, 'tests', 'selftests')
                if not os.path.exists(test_root):
                    if hasattr(os, 'symlink'):
                        os.symlink(files_dir, test_root)
                    else:
                        shutil.copytree(files_dir, test_root)

        elif 'GECKO_INSTALLER_URL' in os.environ:
            # The mochitest tests will run regardless of whether a build exists or not.
            # In a local environment, they should simply be skipped if setup fails. But
            # in automation, we'll need to make sure an error is propagated up.
            pytest.fail(HARNESS_ROOT_NOT_FOUND)
        else:
            # Tests will be marked skipped by the calls to pytest.importorskip() below.
            # We are purposefully not failing here because running |mach python-test|
            # without a build is a perfectly valid use case.
            pass
    return inner


@pytest.fixture(scope='session')
def binary():
    """Return a Firefox binary"""
    try:
        return build.get_binary_path()
    except:
        pass

    app = 'firefox'
    bindir = os.path.join(os.environ['PYTHON_TEST_TMP'], app)
    if os.path.isdir(bindir):
        try:
            return mozinstall.get_binary(bindir, app_name=app)
        except:
            pass

    if 'GECKO_INSTALLER_URL' in os.environ:
        bindir = mozinstall.install(
            os.environ['GECKO_INSTALLER_URL'], os.environ['PYTHON_TEST_TMP'])
        return mozinstall.get_binary(bindir, app_name='firefox')



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import os
import unittest
from mozunit import main
from tempfile import mkdtemp
from shutil import rmtree

import mozpack.path as mozpath
from mozbuild.backend.configenvironment import PartialConfigEnvironment

config = {
    'defines': {
        'MOZ_FOO': '1',
        'MOZ_BAR': '2',
        'MOZ_NON_GLOBAL': '3',
    },
    'substs': {
        'MOZ_SUBST_1': '1',
        'MOZ_SUBST_2': '2',
        'CPP': 'cpp',
    },
    'non_global_defines': [
        'MOZ_NON_GLOBAL',
    ],
}


class TestPartial(unittest.TestCase):
    def setUp(self):
        self._old_env = dict(os.environ)

    def tearDown(self):
        os.environ.clear()
        os.environ.update(self._old_env)

    def _objdir(self):
        objdir = mkdtemp()
        self.addCleanup(rmtree, objdir)
        return objdir

    def test_auto_substs(self):
        '''Test the automatically set values of ACDEFINES, and ALLDEFINES
        '''
        env = PartialConfigEnvironment(self._objdir())
        env.write_vars(config)
        self.assertEqual(env.substs['ACDEFINES'], '-DMOZ_BAR=2 -DMOZ_FOO=1')
        self.assertEqual(env.defines['ALLDEFINES'], {
            'MOZ_BAR': '2',
            'MOZ_FOO': '1',
        })

    def test_remove_subst(self):
        '''Test removing a subst from the config. The file should be overwritten with 'None'
        '''
        env = PartialConfigEnvironment(self._objdir())
        path = mozpath.join(env.topobjdir, 'config.statusd', 'substs', 'MYSUBST')
        myconfig = config.copy()
        env.write_vars(myconfig)
        with self.assertRaises(KeyError):
            x = env.substs['MYSUBST']
        self.assertFalse(os.path.exists(path))

        myconfig['substs']['MYSUBST'] = 'new'
        env.write_vars(myconfig)

        self.assertEqual(env.substs['MYSUBST'], 'new')
        self.assertTrue(os.path.exists(path))

        del myconfig['substs']['MYSUBST']
        env.write_vars(myconfig)
        with self.assertRaises(KeyError):
            x = env.substs['MYSUBST']
        # Now that the subst is gone, the file still needs to be present so that
        # make can update dependencies correctly. Overwriting the file with
        # 'None' is the same as deleting it as far as the
        # PartialConfigEnvironment is concerned, but make can't track a
        # dependency on a file that doesn't exist.
        self.assertTrue(os.path.exists(path))

    def _assert_deps(self, env, deps):
        deps = sorted(['$(wildcard %s)' % (mozpath.join(env.topobjdir, 'config.statusd', d)) for d in deps])
        self.assertEqual(sorted(env.get_dependencies()), deps)

    def test_dependencies(self):
        '''Test getting dependencies on defines and substs.
        '''
        env = PartialConfigEnvironment(self._objdir())
        env.write_vars(config)
        self._assert_deps(env, [])

        self.assertEqual(env.defines['MOZ_FOO'], '1')
        self._assert_deps(env, ['defines/MOZ_FOO'])

        self.assertEqual(env.defines['MOZ_BAR'], '2')
        self._assert_deps(env, ['defines/MOZ_FOO', 'defines/MOZ_BAR'])

        # Getting a define again shouldn't add a redundant dependency
        self.assertEqual(env.defines['MOZ_FOO'], '1')
        self._assert_deps(env, ['defines/MOZ_FOO', 'defines/MOZ_BAR'])

        self.assertEqual(env.substs['MOZ_SUBST_1'], '1')
        self._assert_deps(env, ['defines/MOZ_FOO', 'defines/MOZ_BAR', 'substs/MOZ_SUBST_1'])

        with self.assertRaises(KeyError):
            x = env.substs['NON_EXISTENT']
        self._assert_deps(env, ['defines/MOZ_FOO', 'defines/MOZ_BAR', 'substs/MOZ_SUBST_1', 'substs/NON_EXISTENT'])
        self.assertEqual(env.substs.get('NON_EXISTENT'), None)

    def test_set_subst(self):
        '''Test setting a subst
        '''
        env = PartialConfigEnvironment(self._objdir())
        env.write_vars(config)

        self.assertEqual(env.substs['MOZ_SUBST_1'], '1')
        env.substs['MOZ_SUBST_1'] = 'updated'
        self.assertEqual(env.substs['MOZ_SUBST_1'], 'updated')

        # A new environment should pull the result from the file again.
        newenv = PartialConfigEnvironment(env.topobjdir)
        self.assertEqual(newenv.substs['MOZ_SUBST_1'], '1')

    def test_env_override(self):
        '''Test overriding a subst with an environment variable
        '''
        env = PartialConfigEnvironment(self._objdir())
        env.write_vars(config)

        self.assertEqual(env.substs['MOZ_SUBST_1'], '1')
        self.assertEqual(env.substs['CPP'], 'cpp')

        # Reset the environment and set some environment variables.
        env = PartialConfigEnvironment(env.topobjdir)
        os.environ['MOZ_SUBST_1'] = 'subst 1 environ'
        os.environ['CPP'] = 'cpp environ'

        # The MOZ_SUBST_1 should be overridden by the environment, while CPP is
        # a special variable and should not.
        self.assertEqual(env.substs['MOZ_SUBST_1'], 'subst 1 environ')
        self.assertEqual(env.substs['CPP'], 'cpp')

    def test_update(self):
        '''Test calling update on the substs or defines pseudo dicts
        '''
        env = PartialConfigEnvironment(self._objdir())
        env.write_vars(config)

        mysubsts = {'NEW': 'new'}
        mysubsts.update(env.substs.iteritems())
        self.assertEqual(mysubsts['NEW'], 'new')
        self.assertEqual(mysubsts['CPP'], 'cpp')

        mydefines = {'DEBUG': '1'}
        mydefines.update(env.defines.iteritems())
        self.assertEqual(mydefines['DEBUG'], '1')
        self.assertEqual(mydefines['MOZ_FOO'], '1')

if __name__ == "__main__":
    main()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import unittest

from compare_locales import compare, paths
from cPickle import loads, dumps


class TestTree(unittest.TestCase):
    '''Test the Tree utility class

    Tree value classes need to be in-place editable
    '''

    def test_empty_dict(self):
        tree = compare.Tree(dict)
        self.assertEqual(list(tree.getContent()), [])
        self.assertDictEqual(
            tree.toJSON(),
            {}
        )

    def test_disjoint_dict(self):
        tree = compare.Tree(dict)
        tree['one/entry']['leaf'] = 1
        tree['two/other']['leaf'] = 2
        self.assertEqual(
            list(tree.getContent()),
            [
                (0, 'key', ('one', 'entry')),
                (1, 'value', {'leaf': 1}),
                (0, 'key', ('two', 'other')),
                (1, 'value', {'leaf': 2})
            ]
        )
        self.assertDictEqual(
            tree.toJSON(),
            {
                'one/entry':
                    {'leaf': 1},
                'two/other':
                    {'leaf': 2}
            }
        )
        self.assertMultiLineEqual(
            str(tree),
            '''\
one/entry
    {'leaf': 1}
two/other
    {'leaf': 2}\
'''
        )

    def test_overlapping_dict(self):
        tree = compare.Tree(dict)
        tree['one/entry']['leaf'] = 1
        tree['one/other']['leaf'] = 2
        self.assertEqual(
            list(tree.getContent()),
            [
                (0, 'key', ('one',)),
                (1, 'key', ('entry',)),
                (2, 'value', {'leaf': 1}),
                (1, 'key', ('other',)),
                (2, 'value', {'leaf': 2})
            ]
        )
        self.assertDictEqual(
            tree.toJSON(),
            {
                'one': {
                    'entry':
                        {'leaf': 1},
                    'other':
                        {'leaf': 2}
                }
            }
        )


class TestObserver(unittest.TestCase):
    def test_simple(self):
        obs = compare.Observer()
        f = paths.File('/some/real/sub/path', 'sub/path', locale='de')
        obs.notify('missingEntity', f, 'one')
        obs.notify('missingEntity', f, 'two')
        obs.updateStats(f, {'missing': 15})
        self.assertDictEqual(obs.toJSON(), {
            'summary': {
                'de': {
                    'missing': 15
                }
            },
            'details': {
                'de/sub/path':
                    [{'missingEntity': 'one'},
                     {'missingEntity': 'two'}]
            }
        })
        clone = loads(dumps(obs))
        self.assertDictEqual(clone.summary, obs.summary)
        self.assertDictEqual(clone.details.toJSON(), obs.details.toJSON())
        self.assertIsNone(clone.file_stats)

    def test_module(self):
        obs = compare.Observer(file_stats=True)
        f = paths.File('/some/real/sub/path', 'path',
                       module='sub', locale='de')
        obs.notify('missingEntity', f, 'one')
        obs.notify('obsoleteEntity', f, 'bar')
        obs.notify('missingEntity', f, 'two')
        obs.updateStats(f, {'missing': 15})
        self.assertDictEqual(obs.toJSON(), {
            'summary': {
                'de': {
                    'missing': 15
                }
            },
            'details': {
                'de/sub/path':
                    [
                     {'missingEntity': 'one'},
                     {'obsoleteEntity': 'bar'},
                     {'missingEntity': 'two'},
                    ]
            }
        })
        self.assertDictEqual(obs.file_stats, {
            'de': {
                'sub/path': {
                    'missing': 15
                }
            }
        })
        self.assertEqual(obs.serialize(), '''\
de/sub/path
    +one
    -bar
    +two
de:
missing: 15
0% of entries changed''')
        clone = loads(dumps(obs))
        self.assertDictEqual(clone.summary, obs.summary)
        self.assertDictEqual(clone.details.toJSON(), obs.details.toJSON())
        self.assertDictEqual(clone.file_stats, obs.file_stats)

    def test_file_stats(self):
        obs = compare.Observer(file_stats=True)
        f = paths.File('/some/real/sub/path', 'sub/path', locale='de')
        obs.notify('missingEntity', f, 'one')
        obs.notify('missingEntity', f, 'two')
        obs.updateStats(f, {'missing': 15})
        self.assertDictEqual(obs.toJSON(), {
            'summary': {
                'de': {
                    'missing': 15
                }
            },
            'details': {
                'de/sub/path':
                    [
                     {'missingEntity': 'one'},
                     {'missingEntity': 'two'},
                    ]
            }
        })
        self.assertDictEqual(obs.file_stats, {
            'de': {
                'sub/path': {
                    'missing': 15
                }
            }
        })
        clone = loads(dumps(obs))
        self.assertDictEqual(clone.summary, obs.summary)
        self.assertDictEqual(clone.details.toJSON(), obs.details.toJSON())
        self.assertDictEqual(clone.file_stats, obs.file_stats)


class TestAddRemove(unittest.TestCase):

    def _test(self, left, right, ref_actions):
        ar = compare.AddRemove()
        ar.set_left(left)
        ar.set_right(right)
        actions = list(ar)
        self.assertListEqual(actions, ref_actions)

    def test_equal(self):
        self._test(['z', 'a', 'p'], ['z', 'a', 'p'], [
                ('equal', 'z'),
                ('equal', 'a'),
                ('equal', 'p'),
            ])

    def test_add_start(self):
        self._test(['a', 'p'], ['z', 'a', 'p'], [
                ('add', 'z'),
                ('equal', 'a'),
                ('equal', 'p'),
            ])

    def test_add_middle(self):
        self._test(['z', 'p'], ['z', 'a', 'p'], [
                ('equal', 'z'),
                ('add', 'a'),
                ('equal', 'p'),
            ])

    def test_add_end(self):
        self._test(['z', 'a'], ['z', 'a', 'p'], [
                ('equal', 'z'),
                ('equal', 'a'),
                ('add', 'p'),
            ])

    def test_delete_start(self):
        self._test(['z', 'a', 'p'], ['a', 'p'], [
                ('delete', 'z'),
                ('equal', 'a'),
                ('equal', 'p'),
            ])

    def test_delete_middle(self):
        self._test(['z', 'a', 'p'], ['z', 'p'], [
                ('equal', 'z'),
                ('delete', 'a'),
                ('equal', 'p'),
            ])

    def test_delete_end(self):
        self._test(['z', 'a', 'p'], ['z', 'a'], [
                ('equal', 'z'),
                ('equal', 'a'),
                ('delete', 'p'),
            ])

    def test_replace_start(self):
        self._test(['b', 'a', 'p'], ['z', 'a', 'p'], [
                ('add', 'z'),
                ('delete', 'b'),
                ('equal', 'a'),
                ('equal', 'p'),
            ])

    def test_replace_middle(self):
        self._test(['z', 'b', 'p'], ['z', 'a', 'p'], [
                ('equal', 'z'),
                ('add', 'a'),
                ('delete', 'b'),
                ('equal', 'p'),
            ])

    def test_replace_end(self):
        self._test(['z', 'a', 'b'], ['z', 'a', 'p'], [
                ('equal', 'z'),
                ('equal', 'a'),
                ('add', 'p'),
                ('delete', 'b'),
            ])



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


from __future__ import absolute_import

import time

from mozterm import Terminal

from . import base
from .process import strstatus
from ..handlers import SummaryHandler


def format_seconds(total):
    """Format number of seconds to MM:SS.DD form."""
    minutes, seconds = divmod(total, 60)
    return '%2d:%05.2f' % (minutes, seconds)


class MachFormatter(base.BaseFormatter):

    def __init__(self, start_time=None, write_interval=False, write_times=True,
                 terminal=None, disable_colors=False, summary_on_shutdown=False, **kwargs):
        super(MachFormatter, self).__init__(**kwargs)

        if start_time is None:
            start_time = time.time()
        start_time = int(start_time * 1000)
        self.start_time = start_time
        self.write_interval = write_interval
        self.write_times = write_times
        self.status_buffer = {}
        self.has_unexpected = {}
        self.last_time = None
        self.term = Terminal(disable_styling=disable_colors)
        self.verbose = False
        self._known_pids = set()

        self.summary = SummaryHandler()
        self.summary_on_shutdown = summary_on_shutdown

    def __call__(self, data):
        self.summary(data)

        s = super(MachFormatter, self).__call__(data)
        if s is None:
            return

        time = self.term.dim_blue(format_seconds(self._time(data)))
        return "%s %s\n" % (time, s)

    def _get_test_id(self, data):
        test_id = data.get("test")
        if isinstance(test_id, list):
            test_id = tuple(test_id)
        return test_id

    def _get_file_name(self, test_id):
        if isinstance(test_id, (str, unicode)):
            return test_id

        if isinstance(test_id, tuple):
            return "".join(test_id)

        assert False, "unexpected test_id"

    def suite_start(self, data):
        num_tests = reduce(lambda x, y: x + len(y), data['tests'].itervalues(), 0)
        action = self.term.yellow(data['action'].upper())
        name = ""
        if 'name' in data:
            name = " %s -" % (data['name'],)
        return "%s:%s running %i tests" % (action, name, num_tests)

    def suite_end(self, data):
        action = self.term.yellow(data['action'].upper())
        rv = [action]
        if not self.summary_on_shutdown:
            rv.append(self._format_suite_summary(self.summary.current_suite, self.summary.current))
        return "\n".join(rv)

    def _format_expected(self, status, expected):
        color = self.term.red
        if expected in ("PASS", "OK"):
            return color(status)

        return color("%s expected %s" % (status, expected))

    def _format_suite_summary(self, suite, summary):
        count = summary['counts']
        logs = summary['unexpected_logs']

        rv = ["", self.term.yellow(suite), self.term.yellow("~" * len(suite))]

        # Format check counts
        checks = self.summary.aggregate('count', count)
        rv.append("Ran {} checks ({})".format(sum(checks.values()),
                  ', '.join(['{} {}s'.format(v, k) for k, v in checks.items() if v])))

        # Format expected counts
        checks = self.summary.aggregate('expected', count, include_skip=False)
        rv.append("Expected results: {}".format(sum(checks.values())))

        # Format skip counts
        skip_tests = count["test"]["expected"]["skip"]
        skip_subtests = count["subtest"]["expected"]["skip"]
        if skip_tests:
            skipped = "Skipped: {} tests".format(skip_tests)
            if skip_subtests:
                skipped = "{}, {} subtests".format(skipped, skip_subtests)
            rv.append(skipped)

        # Format unexpected counts
        checks = self.summary.aggregate('unexpected', count)
        unexpected_count = sum(checks.values())
        if unexpected_count:
            rv.append("Unexpected results: {}".format(unexpected_count))
            for key in ('test', 'subtest', 'assert'):
                if not count[key]['unexpected']:
                    continue
                status_str = ", ".join(["{} {}".format(n, s)
                                        for s, n in count[key]['unexpected'].items()])
                rv.append("  {}: {} ({})".format(
                          key, sum(count[key]['unexpected'].values()), status_str))

        # Format status
        if not any(count[key]["unexpected"] for key in ('test', 'subtest', 'assert')):
            rv.append(self.term.green("OK"))
        else:
            heading = "Unexpected Logs"
            rv.extend(["", heading, "-" * len(heading)])
            if count['subtest']['count']:
                for test_id, results in logs.items():
                    test = self._get_file_name(test_id)
                    rv.append(test)
                    for data in results:
                        name = data.get("subtest", "[Parent]")
                        rv.append("  %s %s" % (self._format_expected(
                                             data["status"], data["expected"]), name))
            else:
                for test_id, results in logs.items():
                    test = self._get_file_name(test_id)
                    rv.append(test)
                    assert len(results) == 1
                    data = results[0]
                    assert "subtest" not in data
                    rv.append("  %s %s" % (self._format_expected(
                                           data["status"], data["expected"]), test))

        return "\n".join(rv)

    def test_start(self, data):
        action = self.term.yellow(data['action'].upper())
        return "%s: %s" % (action, self._get_test_id(data))

    def test_end(self, data):
        subtests = self._get_subtest_data(data)

        message = data.get("message", "")
        if "stack" in data:
            stack = data["stack"]
            if stack and stack[-1] != "\n":
                stack += "\n"
            message = stack + message

        if "expected" in data:
            parent_unexpected = True
            expected_str = ", expected %s" % data["expected"]
        else:
            parent_unexpected = False
            expected_str = ""

        test = self._get_test_id(data)

        # Reset the counts to 0
        self.status_buffer[test] = {"count": 0, "unexpected": 0, "pass": 0}
        self.has_unexpected[test] = bool(subtests['unexpected'])

        if subtests["count"] != 0:
            rv = "Test %s%s. Subtests passed %i/%i. Unexpected %s" % (
                data["status"], expected_str, subtests["pass"], subtests["count"],
                subtests['unexpected'])
        else:
            rv = "%s%s" % (data["status"], expected_str)

        unexpected = self.summary.current["unexpected_logs"].get(data["test"])
        if unexpected:
            rv += "\n"
            if len(unexpected) == 1 and parent_unexpected:
                rv += "%s" % unexpected[0].get("message", "")
            else:
                for data in unexpected:
                    name = data.get("subtest", "[Parent]")
                    expected_str = "Expected %s, got %s" % (data["expected"], data["status"])
                    rv += "%s\n" % (
                        "\n".join([name, "-" * len(name), expected_str, data.get("message", "")]))
                rv = rv[:-1]

        if "expected" not in data and not bool(subtests['unexpected']):
            color = self.term.green
        else:
            color = self.term.red
        action = color(data['action'].upper())
        return "%s: %s" % (action, rv)

    def valgrind_error(self, data):
        rv = " " + data['primary'] + "\n"
        for line in data['secondary']:
            rv = rv + line + "\n"

        return rv

    def test_status(self, data):
        test = self._get_test_id(data)
        if test not in self.status_buffer:
            self.status_buffer[test] = {"count": 0, "unexpected": 0, "pass": 0}
        self.status_buffer[test]["count"] += 1

        message = data.get("message", "")
        if "stack" in data:
            if message:
                message += "\n"
            message += data["stack"]

        if data["status"] == "PASS":
            self.status_buffer[test]["pass"] += 1

        rv = None
        status, subtest = data["status"], data["subtest"]
        unexpected = "expected" in data
        if self.verbose:
            status = (self.term.red if unexpected else self.term.green)(status)
            rv = " ".join([subtest, status, message])

        if unexpected:
            self.status_buffer[test]["unexpected"] += 1
        if rv:
            action = self.term.yellow(data['action'].upper())
            return "%s: %s" % (action, rv)

    def assertion_count(self, data):
        if data["min_expected"] <= data["count"] <= data["max_expected"]:
            return

        if data["min_expected"] != data["max_expected"]:
            expected = "%i to %i" % (data["min_expected"],
                                     data["max_expected"])
        else:
            expected = "%i" % data["min_expected"]

        action = self.term.red("ASSERT")
        return "%s: Assertion count %i, expected %i assertions\n" % (
                action, data["count"], expected)

    def process_output(self, data):
        rv = []

        pid = data['process']
        if pid.isdigit():
            pid = 'pid:%s' % pid
        pid = self.term.dim_cyan(pid)

        if "command" in data and data["process"] not in self._known_pids:
            self._known_pids.add(data["process"])
            rv.append('%s Full command: %s' % (pid, data["command"]))

        rv.append('%s %s' % (pid, data["data"]))
        return "\n".join(rv)

    def crash(self, data):
        test = self._get_test_id(data)

        if data.get("stackwalk_returncode", 0) != 0 and not data.get("stackwalk_stderr"):
            success = True
        else:
            success = False

        rv = ["pid:%s. Test:%s. Minidump anaylsed:%s. Signature:[%s]" %
              (data.get("pid", None), test, success, data["signature"])]

        if data.get("minidump_path"):
            rv.append("Crash dump filename: %s" % data["minidump_path"])

        if data.get("stackwalk_returncode", 0) != 0:
            rv.append("minidump_stackwalk exited with return code %d" %
                      data["stackwalk_returncode"])

        if data.get("stackwalk_stderr"):
            rv.append("stderr from minidump_stackwalk:")
            rv.append(data["stackwalk_stderr"])
        elif data.get("stackwalk_stdout"):
            rv.append(data["stackwalk_stdout"])

        if data.get("stackwalk_errors"):
            rv.extend(data.get("stackwalk_errors"))

        rv = "\n".join(rv)
        if not rv[-1] == "\n":
            rv += "\n"

        action = self.term.red(data['action'].upper())
        return "%s: %s" % (action, rv)

    def process_start(self, data):
        rv = "Started process `%s`" % data['process']
        desc = data.get('command')
        if desc:
            rv = '%s (%s)' % (rv, desc)
        return rv

    def process_exit(self, data):
        return "%s: %s" % (data['process'], strstatus(data['exitcode']))

    def log(self, data):
        level = data.get("level").upper()

        if level in ("CRITICAL", "ERROR"):
            level = self.term.red(level)
        elif level == "WARNING":
            level = self.term.yellow(level)
        elif level == "INFO":
            level = self.term.blue(level)

        if data.get('component'):
            rv = " ".join([data["component"], level, data["message"]])
        else:
            rv = "%s %s" % (level, data["message"])

        if "stack" in data:
            rv += "\n%s" % data["stack"]

        return rv

    def lint(self, data):
        fmt = "{path}  {c1}{lineno}{column}  {c2}{level}{normal}  {message}" \
              "  {c1}{rule}({linter}){normal}"
        message = fmt.format(
            path=data["path"],
            normal=self.term.normal,
            c1=self.term.grey,
            c2=self.term.red if data["level"] == 'error' else self.term.yellow,
            lineno=str(data["lineno"]),
            column=(":" + str(data["column"])) if data.get("column") else "",
            level=data["level"],
            message=data["message"],
            rule='{} '.format(data["rule"]) if data.get("rule") else "",
            linter=data["linter"].lower() if data.get("linter") else "",
        )

        return message

    def shutdown(self, data):
        if not self.summary_on_shutdown:
            return

        heading = "Overall Summary"
        rv = ["", self.term.bold_yellow(heading), self.term.bold_yellow("=" * len(heading))]
        for suite, summary in self.summary:
            rv.append(self._format_suite_summary(suite, summary))
        return "\n".join(rv)

    def _get_subtest_data(self, data):
        test = self._get_test_id(data)
        return self.status_buffer.get(test, {"count": 0, "unexpected": 0, "pass": 0})

    def _time(self, data):
        entry_time = data["time"]
        if self.write_interval and self.last_time is not None:
            t = entry_time - self.last_time
            self.last_time = entry_time
        else:
            t = entry_time - self.start_time

        return t / 1000.



# -*- coding: utf-8 -*-

# Any copyright is dedicated to the Public Domain.
# http://creativecommons.org/publicdomain/zero/1.0/

import os

import mozunit

from unittest import expectedFailure

from mozpack.copier import (
    FileCopier,
    FileRegistry,
)
from mozpack.manifests import (
    InstallManifest,
    UnreadableInstallManifest,
)
from mozpack.test.test_files import TestWithTmpDir

import mozbuild.action.process_install_manifest as process_install_manifest


class TestGenerateManifest(TestWithTmpDir):
    """
    Unit tests for process_install_manifest.py.
    """

    def test_process_manifest(self):
        source = self.tmppath('source')
        os.mkdir(source)
        os.mkdir('%s/base' % source)
        os.mkdir('%s/base/foo' % source)
        os.mkdir('%s/base2' % source)

        with open('%s/base/foo/file1' % source, 'a'):
            pass

        with open('%s/base/foo/file2' % source, 'a'):
            pass

        with open('%s/base2/file3' % source, 'a'):
            pass

        m = InstallManifest()
        m.add_pattern_link('%s/base' % source, '**', '')
        m.add_link('%s/base2/file3' % source, 'foo/file3')

        p = self.tmppath('m')
        m.write(path=p)

        dest = self.tmppath('dest')
        track = self.tmppath('track')

        for i in range(2):
            process_install_manifest.process_manifest(dest, [p], track)

            self.assertTrue(os.path.exists(self.tmppath('dest/foo/file1')))
            self.assertTrue(os.path.exists(self.tmppath('dest/foo/file2')))
            self.assertTrue(os.path.exists(self.tmppath('dest/foo/file3')))

        m = InstallManifest()
        m.write(path=p)

        for i in range(2):
            process_install_manifest.process_manifest(dest, [p], track)

            self.assertFalse(os.path.exists(self.tmppath('dest/foo/file1')))
            self.assertFalse(os.path.exists(self.tmppath('dest/foo/file2')))
            self.assertFalse(os.path.exists(self.tmppath('dest/foo/file3')))

if __name__ == '__main__':
    mozunit.main()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import mozlog
import time

import pytest


def pytest_addoption(parser):
    # We can't simply use mozlog.commandline.add_logging_group(parser) here because
    # Pytest's parser doesn't have the add_argument_group method Mozlog expects.
    group = parser.getgroup('mozlog')

    for name, (_class, _help) in mozlog.commandline.log_formatters.iteritems():
        group.addoption('--log-{0}'.format(name), action='append', help=_help)

    formatter_options = mozlog.commandline.fmt_options.iteritems()
    for name, (_class, _help, formatters, action) in formatter_options:
        for formatter in formatters:
            if formatter in mozlog.commandline.log_formatters:
                group.addoption(
                    '--log-{0}-{1}'.format(formatter, name),
                    action=action,
                    help=_help)


def pytest_configure(config):
    # If using pytest-xdist for parallelization, only register plugin on master process
    if not hasattr(config, 'slaveinput'):
        config.pluginmanager.register(MozLog())


class MozLog(object):

    def __init__(self):
        self._started = False
        self.results = {}
        self.start_time = int(time.time() * 1000)  # in ms for Mozlog compatibility

    def _log_suite_start(self, tests):
        if not self._started:
            # As this is called for each node when using pytest-xdist, we want
            # to avoid logging multiple suite_start messages.
            self.logger.suite_start(
                tests=tests,
                time=self.start_time,
                run_info=self.run_info)
            self._started = True

    def pytest_configure(self, config):
        mozlog.commandline.setup_logging('pytest', config.known_args_namespace,
                                         defaults={}, allow_unused_options=True)
        self.logger = mozlog.get_default_logger(component='pytest')

    def pytest_sessionstart(self, session):
        '''Called before test collection; records suite start time to log later'''
        self.start_time = int(time.time() * 1000)  # in ms for Mozlog compatibility
        self.run_info = getattr(session.config, '_metadata', None)

    def pytest_collection_finish(self, session):
        '''Called after test collection is completed, just before tests are run (suite start)'''
        self._log_suite_start([item.nodeid for item in session.items])

    @pytest.mark.optionalhook
    def pytest_xdist_node_collection_finished(self, node, ids):
        '''Called after each pytest-xdist node collection is completed'''
        self._log_suite_start(ids)

    def pytest_sessionfinish(self, session, exitstatus):
        self.logger.suite_end()

    def pytest_runtest_logstart(self, nodeid, location):
        self.logger.test_start(test=nodeid)

    def pytest_runtest_logreport(self, report):
        '''Called 3 times per test (setup, call, teardown), indicated by report.when'''
        test = report.nodeid
        status = expected = 'PASS'
        message = stack = None
        if hasattr(report, 'wasxfail'):
            expected = 'FAIL'
        if report.failed or report.outcome == 'rerun':
            status = 'FAIL' if report.when == 'call' else 'ERROR'
        if report.skipped:
            status = 'SKIP' if not hasattr(report, 'wasxfail') else 'FAIL'
        if report.longrepr is not None:
            longrepr = report.longrepr
            if isinstance(longrepr, basestring):
                # When using pytest-xdist, longrepr is serialised as a str
                message = stack = longrepr
                if longrepr.startswith('[XPASS(strict)]'):
                    # Strict expected failures have an outcome of failed when
                    # they unexpectedly pass.
                    expected, status = ('FAIL', 'PASS')
            elif hasattr(longrepr, "reprcrash"):
                # For failures, longrepr is a ReprExceptionInfo
                crash = longrepr.reprcrash
                message = "{0} (line {1})".format(crash.message, crash.lineno)
                stack = longrepr.reprtraceback
            elif hasattr(longrepr, "errorstring"):
                message = longrepr.errorstring
                stack = longrepr.errorstring
            elif hasattr(longrepr, "__getitem__") and len(longrepr) == 3:
                # For skips, longrepr is a tuple of (file, lineno, reason)
                message = report.longrepr[-1]
            else:
                raise ValueError("Unable to convert longrepr to message:\ntype %s\nfields:" %
                                 (longrepr.__class__, dir(longrepr)))
        if status != expected or expected != 'PASS':
            self.results[test] = (status, expected, message, stack)
        if report.outcome == 'rerun' or report.when == 'teardown':
            defaults = ('PASS', 'PASS', None, None)
            status, expected, message, stack = self.results.get(test, defaults)
            self.logger.test_end(test=test, status=status, expected=expected,
                                 message=message, stack=stack)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

###
# This script generates a web manifest JSON file based on the xpi-stage
# directory structure. It extracts the data from defines.inc files from
# the locale directory, chrome registry entries and other information
# necessary to produce the complete manifest file for a language pack.
###
from __future__ import absolute_import

import argparse
import sys
import os
import json
import io
import datetime
import requests
import mozversioncontrol
import mozpack.path as mozpath
from mozpack.chrome.manifest import (
    Manifest,
    ManifestLocale,
    parse_manifest,
)
from mozbuild.preprocessor import Preprocessor


def write_file(path, content):
    with io.open(path, 'w', encoding='utf-8') as out:
        out.write(content + '\n')


pushlog_api_url = "{0}/json-rev/{1}"


###
# Retrievers a UTC datetime of the push for the current commit
# from a mercurial clone directory.
#
# Args:
#    path (str) - path to a directory
#
# Returns:
#    (datetime) - a datetime object
#
# Example:
#    dt = get_dt_from_hg("/var/vcs/l10n-central/pl")
#    dt == datetime(2017, 10, 11, 23, 31, 54, 0)
###
def get_dt_from_hg(path):
    with mozversioncontrol.get_repository_object(path=path) as repo:
        phase = repo._run_in_client(["log", "-r", ".", "-T" "{phase}"])
        if phase.strip() != "public":
            return datetime.datetime.utcnow()
        repo_url = repo._run_in_client(["paths", "default"])
        repo_url = repo_url.strip().replace("ssh://", "https://")
        repo_url = repo_url.replace("hg://", "https://")
        cs = repo._run_in_client(["log", "-r", ".", "-T" "{node}"])

    url = pushlog_api_url.format(repo_url, cs)
    session = requests.Session()
    try:
        response = session.get(url)
    except Exception as e:
        msg = "Failed to retrieve push timestamp using {}\nError: {}".format(url, e)
        raise Exception(msg)

    data = response.json()

    date = data['pushdate'][0]

    return datetime.datetime.utcfromtimestamp(date)


###
# Generates timestamp for a locale based on its path.
# If possible, will use the commit timestamp from HG repository,
# and if that fails, will generate the timestamp for `now`.
#
# The timestamp format is "{year}{month}{day}{hour}{minute}{second}" and
# the datetime stored in it is using UTC timezone.
#
# Args:
#    path (str) - path to the locale directory
#
# Returns:
#    (str) - a timestamp string
#
# Example:
#    ts = get_timestamp_for_locale("/var/vcs/l10n-central/pl")
#    ts == "20170914215617"
###
def get_timestamp_for_locale(path):
    dt = None
    if os.path.isdir(os.path.join(path, '.hg')):
        dt = get_dt_from_hg(path)

    if dt is None:
        dt = datetime.datetime.utcnow()

    dt = dt.replace(microsecond=0)
    return dt.strftime("%Y%m%d%H%M%S")


###
# Parses multiple defines files into a single key-value pair object.
#
# Args:
#    paths (str) - a comma separated list of paths to defines files
#
# Returns:
#    (dict) - a key-value dict with defines
#
# Example:
#    res = parse_defines('./toolkit/defines.inc,./browser/defines.inc')
#    res == {
#        'MOZ_LANG_TITLE': 'Polski',
#        'MOZ_LANGPACK_CREATOR': 'Aviary.pl',
#        'MOZ_LANGPACK_CONTRIBUTORS': 'Marek Stepien, Marek Wawoczny'
#    }
###
def parse_defines(paths):
    pp = Preprocessor()
    for path in paths:
        pp.do_include(path)

    return pp.context


###
# Converts the list of contributors from the old RDF based list
# of entries, into a comma separated list.
#
# Args:
#    str (str) - a string with an RDF list of contributors entries
#
# Returns:
#    (str) - a comma separated list of contributors
#
# Example:
#    s = convert_contributors('
#        <em:contributor>Marek Wawoczny</em:contributor>
#        <em:contributor>Marek Stepien</em:contributor>
#    ')
#    s == 'Marek Wawoczny, Marek Stepien'
###
def convert_contributors(str):
    str = str.replace('<em:contributor>', '')
    tokens = str.split('</em:contributor>')
    tokens = map(lambda t: t.strip(), tokens)
    tokens = filter(lambda t: t != '', tokens)
    return ', '.join(tokens)


###
# Build the manifest author string based on the author string
# and optionally adding the list of contributors, if provided.
#
# Args:
#    author (str)       - a string with the name of the author
#    contributors (str) - RDF based list of contributors from a chrome manifest
#
# Returns:
#    (str) - a string to be placed in the author field of the manifest.json
#
# Example:
#    s = build_author_string(
#    'Aviary.pl',
#    '
#        <em:contributor>Marek Wawoczny</em:contributor>
#        <em:contributor>Marek Stepien</em:contributor>
#    ')
#    s == 'Aviary.pl (contributors: Marek Wawoczny, Marek Stepien)'
###
def build_author_string(author, contributors):
    contrib = convert_contributors(contributors)
    if len(contrib) == 0:
        return author
    return '{0} (contributors: {1})'.format(author, contrib)


##
# Converts the list of chrome manifest entry flags to the list of platforms
# for the langpack manifest.
#
# The list of result platforms is taken from AppConstants.platform.
#
# Args:
#    flags (FlagList) - a list of Chrome Manifest entry flags
#
# Returns:
#    (list) - a list of platform the entry applies to
#
# Example:
#    str(flags) == "os==MacOS os==Windows"
#    platforms = convert_entry_flags_to_platform_codes(flags)
#    platforms == ['mac', 'win']
#
# The method supports only `os` flag name and equality operator.
# It will throw if tried with other flags or operators.
###
def convert_entry_flags_to_platform_codes(flags):
    if not flags:
        return None

    ret = []
    for key in flags:
        if key != 'os':
            raise Exception('Unknown flag name')

        for value in flags[key].values:
            if value[0] != '==':
                raise Exception('Inequality flag cannot be converted')

            if value[1] == 'Android':
                ret.append('android')
            elif value[1] == 'LikeUnix':
                ret.append('linux')
            elif value[1] == 'Darwin':
                ret.append('macosx')
            elif value[1] == 'WINNT':
                ret.append('win')
            else:
                raise Exception('Unknown flag value {0}'.format(value[1]))

    return ret


###
# Recursively parse a chrome manifest file appending new entries
# to the result list
#
# The function can handle two entry types: 'locale' and 'manifest'
#
# Args:
#    path           (str)  - a path to a chrome manifest
#    base_path      (str)  - a path to the base directory all chrome registry
#                            entries will be relative to
#    chrome_entries (list) - a list to which entries will be appended to
#
# Example:
#
#    chrome_entries = {}
#    parse_manifest('./chrome.manifest', './', chrome_entries)
#
#    chrome_entries == [
#        {
#            'type': 'locale',
#            'alias': 'devtools',
#            'locale': 'pl',
#            'platforms': null,
#            'path': 'chrome/pl/locale/pl/devtools/'
#        },
#        {
#            'type': 'locale',
#            'alias': 'autoconfig',
#            'locale': 'pl',
#            'platforms': ['win', 'mac'],
#            'path': 'chrome/pl/locale/pl/autoconfig/'
#        },
#    ]
###
def parse_chrome_manifest(path, base_path, chrome_entries):
    for entry in parse_manifest(None, path):
        if isinstance(entry, Manifest):
            parse_chrome_manifest(
                os.path.join(os.path.dirname(path), entry.relpath),
                base_path,
                chrome_entries
            )
        elif isinstance(entry, ManifestLocale):
            entry_path = os.path.join(
                os.path.relpath(
                    os.path.dirname(path),
                    base_path
                ),
                entry.relpath
            )
            chrome_entries.append({
                'type': 'locale',
                'alias': entry.name,
                'locale': entry.id,
                'platforms': convert_entry_flags_to_platform_codes(entry.flags),
                'path': mozpath.normsep(entry_path)
            })
        else:
            raise Exception('Unknown type {0}'.format(entry.name))


###
# Generates a new web manifest dict with values specific for a language pack.
#
# Args:
#    locstr         (str)  - A string with a comma separated list of locales
#                            for which resources are embedded in the
#                            language pack
#    min_app_ver    (str)  - A minimum version of the application the language
#                            resources are for
#    max_app_ver    (str)  - A maximum version of the application the language
#                            resources are for
#    app_name       (str)  - The name of the application the language
#                            resources are for
#    defines        (dict) - A dictionary of defines entries
#    chrome_entries (dict) - A dictionary of chrome registry entries
#
# Returns:
#    (dict) - a web manifest
#
# Example:
#    manifest = create_webmanifest(
#      'pl',
#      '57.0',
#      '57.0.*',
#      'Firefox',
#      '/var/vcs/l10n-central',
#      {'MOZ_LANG_TITLE': 'Polski'},
#      chrome_entries
#    )
#    manifest == {
#        'languages': {
#            'pl': {
#                'version': '201709121481',
#                'chrome_resources': {
#                    'alert': 'chrome/pl/locale/pl/alert/',
#                    'branding': 'browser/chrome/pl/locale/global/',
#                    'global-platform': {
#                      'macosx': 'chrome/pl/locale/pl/global-platform/mac/',
#                      'win': 'chrome/pl/locale/pl/global-platform/win/',
#                      'linux': 'chrome/pl/locale/pl/global-platform/unix/',
#                      'android': 'chrome/pl/locale/pl/global-platform/unix/',
#                    },
#                    'forms': 'browser/chrome/pl/locale/forms/',
#                    ...
#                }
#            }
#        },
#        'sources': {
#            'browser': {
#                'base_path': 'browser/'
#            }
#        },
#        'applications': {
#            'gecko':  {
#                'strict_min_version': '57.0',
#                'strict_max_version': '57.0.*',
#                'id': 'langpack-pl@mozilla.org',
#            }
#        },
#        'version': '57.0',
#        'name': 'Polski Language Pack',
#        ...
#    }
###
def create_webmanifest(locstr, min_app_ver, max_app_ver, app_name,
                       l10n_basedir, defines, chrome_entries):
    locales = map(lambda loc: loc.strip(), locstr.split(','))
    main_locale = locales[0]

    author = build_author_string(
        defines['MOZ_LANGPACK_CREATOR'],
        defines['MOZ_LANGPACK_CONTRIBUTORS'] if 'MOZ_LANGPACK_CONTRIBUTORS' in defines else ""
    )

    manifest = {
        'langpack_id': main_locale,
        'manifest_version': 2,
        'applications': {
            'gecko': {
                'id': 'langpack-{0}@firefox.mozilla.org'.format(main_locale),
                'strict_min_version': min_app_ver,
                'strict_max_version': max_app_ver,
            }
        },
        'name': '{0} Language Pack'.format(defines['MOZ_LANG_TITLE']),
        'description': 'Language pack for {0} for {1}'.format(app_name, main_locale),
        'version': min_app_ver,
        'languages': {},
        'sources': {
            'browser': {
                'base_path': 'browser/'
            }
        },
        'author': author
    }

    cr = {}
    for entry in chrome_entries:
        if entry['type'] == 'locale':
            platforms = entry['platforms']
            if platforms:
                if entry['alias'] not in cr:
                    cr[entry['alias']] = {}
                for platform in platforms:
                    cr[entry['alias']][platform] = entry['path']
            else:
                assert entry['alias'] not in cr
                cr[entry['alias']] = entry['path']
        else:
            raise Exception('Unknown type {0}'.format(entry['type']))

    for loc in locales:
        manifest['languages'][loc] = {
            'version': get_timestamp_for_locale(os.path.join(l10n_basedir, loc)),
            'chrome_resources': cr
        }

    return json.dumps(manifest, indent=2, ensure_ascii=False, encoding='utf8')


def main(args):
    parser = argparse.ArgumentParser()
    parser.add_argument('--locales',
                        help='List of language codes provided by the langpack')
    parser.add_argument('--min-app-ver',
                        help='Min version of the application the langpack is for')
    parser.add_argument('--max-app-ver',
                        help='Max version of the application the langpack is for')
    parser.add_argument('--app-name',
                        help='Name of the application the langpack is for')
    parser.add_argument('--l10n-basedir',
                        help='Base directory for locales used in the language pack')
    parser.add_argument('--defines', default=[], nargs='+',
                        help='List of defines files to load data from')
    parser.add_argument('--input',
                        help='Langpack directory.')

    args = parser.parse_args(args)

    chrome_entries = []
    parse_chrome_manifest(
        os.path.join(args.input, 'chrome.manifest'), args.input, chrome_entries)

    defines = parse_defines(args.defines)

    res = create_webmanifest(
        args.locales,
        args.min_app_ver,
        args.max_app_ver,
        args.app_name,
        args.l10n_basedir,
        defines,
        chrome_entries
    )
    write_file(os.path.join(args.input, 'manifest.json'), res)


if __name__ == '__main__':
    main(sys.argv[1:])



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import platform

from mozboot.base import BaseBootstrapper
from mozboot.linux_common import StyloInstall


class CentOSFedoraBootstrapper(StyloInstall, BaseBootstrapper):
    def __init__(self, distro, version, dist_id, **kwargs):
        BaseBootstrapper.__init__(self, **kwargs)

        self.distro = distro
        self.version = version
        self.dist_id = dist_id

        self.group_packages = []

        self.packages = [
            'autoconf213',
            'mercurial',
            'nodejs',
            'npm',
            'which',
        ]

        self.browser_group_packages = [
            'GNOME Software Development',
        ]

        self.browser_packages = [
            'alsa-lib-devel',
            'dbus-glib-devel',
            'GConf2-devel',
            'glibc-static',
            'gtk2-devel',  # It is optional in Fedora 20's GNOME Software
                           # Development group.
            'libstdc++-static',
            'libXt-devel',
            'pulseaudio-libs-devel',
            'wireless-tools-devel',
            'yasm',
        ]

        self.mobile_android_packages = [
            'java-1.8.0-openjdk-devel',
            # For downloading the Android SDK and NDK.
            'wget',
        ]

        if self.distro in ('CentOS', 'CentOS Linux'):
            self.group_packages += [
                'Development Tools',
                'Development Libraries',
                'GNOME Software Development',
            ]

            self.packages += [
                'curl-devel',
            ]

            self.browser_packages += [
                'gtk3-devel',
            ]

        elif self.distro == 'Fedora':
            self.group_packages += [
                'C Development Tools and Libraries',
            ]

            self.packages += [
                'python2-devel',
                'redhat-rpm-config',
            ]

            self.browser_packages += [
                'gcc-c++',
                'python-dbus',
            ]

    def install_system_packages(self):
        self.dnf_groupinstall(*self.group_packages)
        self.dnf_install(*self.packages)

    def install_browser_packages(self):
        self.ensure_browser_packages()

    def install_browser_artifact_mode_packages(self):
        self.ensure_browser_packages(artifact_mode=True)

    def install_mobile_android_packages(self):
        self.ensure_mobile_android_packages(artifact_mode=False)

    def install_mobile_android_artifact_mode_packages(self):
        self.ensure_mobile_android_packages(artifact_mode=True)

    def ensure_browser_packages(self, artifact_mode=False):
        # TODO: Figure out what not to install for artifact mode
        self.dnf_groupinstall(*self.browser_group_packages)
        self.dnf_install(*self.browser_packages)

        if self.distro in ('CentOS', 'CentOS Linux'):
            yasm = 'http://pkgs.repoforge.org/yasm/yasm-1.1.0-1.el6.rf.i686.rpm'
            if platform.architecture()[0] == '64bit':
                yasm = 'http://pkgs.repoforge.org/yasm/yasm-1.1.0-1.el6.rf.x86_64.rpm'

            self.run_as_root(['rpm', '-ivh', yasm])

    def ensure_mobile_android_packages(self, artifact_mode=False):
        # Install Android specific packages.
        self.dnf_install(*self.mobile_android_packages)

        from mozboot import android
        android.ensure_android('linux', artifact_mode=artifact_mode,
                               no_interactive=self.no_interactive)

    def suggest_mobile_android_mozconfig(self, artifact_mode=False):
        from mozboot import android
        android.suggest_mozconfig('linux', artifact_mode=artifact_mode)

    def suggest_mobile_android_artifact_mode_mozconfig(self):
        self.suggest_mobile_android_mozconfig(artifact_mode=True)

    def upgrade_mercurial(self, current):
        self.dnf_update('mercurial')



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import

import os
import subprocess

import mozunit
import pytest

from mozversioncontrol import get_repository_object


setup = {
    'hg': [
        """
        echo "foo" > foo
        echo "bar" > bar
        hg init
        hg add *
        hg commit -m "Initial commit"
        """,
        """
        echo "[paths]\ndefault = ../remoterepo" > .hg/hgrc
        """,
        """
        echo "bar" >> bar
        echo "baz" > baz
        hg add baz
        hg rm foo
        """,
        """
        hg commit -m "Remove foo; modify bar; add baz"
        """,
    ],
    'git': [
        """
        echo "foo" > foo
        echo "bar" > bar
        git init
        git add *
        git commit -am "Initial commit"
        """,
        """
        git remote add upstream ../remoterepo
        git fetch upstream
        git branch -u upstream/master
        """,
        """
        echo "bar" >> bar
        echo "baz" > baz
        git add baz
        git rm foo
        """,
        """
        git commit -am "Remove foo; modify bar; add baz"
        """
    ]
}


def shell(cmd):
    subprocess.check_call(cmd, shell=True)


@pytest.yield_fixture(params=['git', 'hg'])
def repo(tmpdir, request):
    vcs = request.param

    # tmpdir and repo are py.path objects
    # http://py.readthedocs.io/en/latest/path.html
    repo = tmpdir.mkdir('repo')
    repo.vcs = vcs

    # This creates a setup iterator. Each time next() is called
    # on it, the next set of instructions will be executed.
    repo.setup = (shell(cmd) for cmd in setup[vcs])

    oldcwd = os.getcwd()
    os.chdir(repo.strpath)

    next(repo.setup)

    repo.copy(tmpdir.join('remoterepo'))

    next(repo.setup)

    yield repo
    os.chdir(oldcwd)


def assert_files(actual, expected):
    assert set(map(os.path.basename, actual)) == set(expected)


def test_workdir_outgoing(repo):
    vcs = get_repository_object(repo.strpath)
    assert vcs.path == repo.strpath

    remotepath = '../remoterepo' if repo.vcs == 'hg' else 'upstream/master'

    next(repo.setup)

    assert_files(vcs.get_changed_files('AM', 'all'), ['bar', 'baz'])
    if repo.vcs == 'git':
        assert_files(vcs.get_changed_files('AM', mode='staged'), ['baz'])
    elif repo.vcs == 'hg':
        assert_files(vcs.get_changed_files('AM', 'staged'), ['bar', 'baz'])
    assert_files(vcs.get_outgoing_files('AM'), [])
    assert_files(vcs.get_outgoing_files('AM', remotepath), [])

    next(repo.setup)

    assert_files(vcs.get_changed_files('AM', 'all'), [])
    assert_files(vcs.get_changed_files('AM', 'staged'), [])
    assert_files(vcs.get_outgoing_files('AM'), ['bar', 'baz'])
    assert_files(vcs.get_outgoing_files('AM', remotepath), ['bar', 'baz'])


if __name__ == '__main__':
    mozunit.main()



from __future__ import unicode_literals
import sys
import json


def to_json(value):
    if isinstance(value, BaseNode):
        return value.to_json()
    if isinstance(value, list):
        return list(map(to_json, value))
    if isinstance(value, tuple):
        return list(map(to_json, value))
    else:
        return value


def from_json(value):
    if isinstance(value, dict):
        cls = getattr(sys.modules[__name__], value['type'])
        args = {
            k: from_json(v)
            for k, v in value.items()
            if k != 'type'
        }
        return cls(**args)
    if isinstance(value, list):
        return list(map(from_json, value))
    else:
        return value


def scalars_equal(node1, node2, ignored_fields):
    """Compare two nodes which are not lists."""

    if type(node1) != type(node2):
        return False

    if isinstance(node1, BaseNode):
        return node1.equals(node2, ignored_fields)

    return node1 == node2


class BaseNode(object):
    """Base class for all Fluent AST nodes.

    All productions described in the ASDL subclass BaseNode, including Span and
    Annotation.  Implements __str__, to_json and traverse.
    """

    def traverse(self, fun):
        """Postorder-traverse this node and apply `fun` to all child nodes.

        Traverse this node depth-first applying `fun` to subnodes and leaves.
        Children are processed before parents (postorder traversal).

        Return a new instance of the node.
        """

        def visit(value):
            """Call `fun` on `value` and its descendants."""
            if isinstance(value, BaseNode):
                return value.traverse(fun)
            if isinstance(value, list):
                return fun(list(map(visit, value)))
            else:
                return fun(value)

        # Use all attributes found on the node as kwargs to the constructor.
        kwargs = vars(self).items()
        node = self.__class__(
            **{name: visit(value) for name, value in kwargs})

        return fun(node)

    def equals(self, other, ignored_fields=['span']):
        """Compare two nodes.

        Nodes are deeply compared on a field by field basis. If possible, False
        is returned early. When comparing attributes and variants in
        SelectExpressions, the order doesn't matter. By default, spans are not
        taken into account.
        """

        self_keys = set(vars(self).keys())
        other_keys = set(vars(other).keys())

        if ignored_fields:
            for key in ignored_fields:
                self_keys.discard(key)
                other_keys.discard(key)

        if self_keys != other_keys:
            return False

        for key in self_keys:
            field1 = getattr(self, key)
            field2 = getattr(other, key)

            # List-typed nodes are compared item-by-item.  When comparing
            # attributes and variants, the order of items doesn't matter.
            if isinstance(field1, list) and isinstance(field2, list):
                if len(field1) != len(field2):
                    return False

                # These functions are used to sort lists of items for when
                # order doesn't matter.  Annotations are also lists but they
                # can't be keyed on any of their fields reliably.
                field_sorting = {
                    'attributes': lambda elem: elem.id.name,
                    'variants': lambda elem: elem.key.name,
                }

                if key in field_sorting:
                    sorting = field_sorting[key]
                    field1 = sorted(field1, key=sorting)
                    field2 = sorted(field2, key=sorting)

                for elem1, elem2 in zip(field1, field2):
                    if not scalars_equal(elem1, elem2, ignored_fields):
                        return False

            elif not scalars_equal(field1, field2, ignored_fields):
                return False

        return True

    def to_json(self):
        obj = {
            name: to_json(value)
            for name, value in vars(self).items()
        }
        obj.update(
            {'type': self.__class__.__name__}
        )
        return obj

    def __str__(self):
        return json.dumps(self.to_json())


class SyntaxNode(BaseNode):
    """Base class for AST nodes which can have Spans."""

    def __init__(self, span=None, **kwargs):
        super(SyntaxNode, self).__init__(**kwargs)
        self.span = span

    def add_span(self, start, end):
        self.span = Span(start, end)


class Resource(SyntaxNode):
    def __init__(self, body=None, **kwargs):
        super(Resource, self).__init__(**kwargs)
        self.body = body or []


class Entry(SyntaxNode):
    def __init__(self, annotations=None, **kwargs):
        super(Entry, self).__init__(**kwargs)
        self.annotations = annotations or []

    def add_annotation(self, annot):
        self.annotations.append(annot)


class Message(Entry):
    def __init__(self, id, value=None, attributes=None,
                 comment=None, **kwargs):
        super(Message, self).__init__(**kwargs)
        self.id = id
        self.value = value
        self.attributes = attributes or []
        self.comment = comment

class Term(Entry):
    def __init__(self, id, value, attributes=None,
                 comment=None, **kwargs):
        super(Term, self).__init__(**kwargs)
        self.id = id
        self.value = value
        self.attributes = attributes or []
        self.comment = comment

class Pattern(SyntaxNode):
    def __init__(self, elements, **kwargs):
        super(Pattern, self).__init__(**kwargs)
        self.elements = elements

class PatternElement(SyntaxNode):
    pass

class TextElement(PatternElement):
    def __init__(self, value, **kwargs):
        super(TextElement, self).__init__(**kwargs)
        self.value = value

class Placeable(PatternElement):
    def __init__(self, expression, **kwargs):
        super(Placeable, self).__init__(**kwargs)
        self.expression = expression

class Expression(SyntaxNode):
    def __init__(self, **kwargs):
        super(Expression, self).__init__(**kwargs)

class StringExpression(Expression):
    def __init__(self, value, **kwargs):
        super(StringExpression, self).__init__(**kwargs)
        self.value = value

class NumberExpression(Expression):
    def __init__(self, value, **kwargs):
        super(NumberExpression, self).__init__(**kwargs)
        self.value = value

class MessageReference(Expression):
    def __init__(self, id, **kwargs):
        super(MessageReference, self).__init__(**kwargs)
        self.id = id

class ExternalArgument(Expression):
    def __init__(self, id, **kwargs):
        super(ExternalArgument, self).__init__(**kwargs)
        self.id = id

class SelectExpression(Expression):
    def __init__(self, expression, variants, **kwargs):
        super(SelectExpression, self).__init__(**kwargs)
        self.expression = expression
        self.variants = variants

class AttributeExpression(Expression):
    def __init__(self, id, name, **kwargs):
        super(AttributeExpression, self).__init__(**kwargs)
        self.id = id
        self.name = name

class VariantExpression(Expression):
    def __init__(self, id, key, **kwargs):
        super(VariantExpression, self).__init__(**kwargs)
        self.id = id
        self.key = key

class CallExpression(Expression):
    def __init__(self, callee, args=None, **kwargs):
        super(CallExpression, self).__init__(**kwargs)
        self.callee = callee
        self.args = args or []

class Attribute(SyntaxNode):
    def __init__(self, id, value, **kwargs):
        super(Attribute, self).__init__(**kwargs)
        self.id = id
        self.value = value

class Variant(SyntaxNode):
    def __init__(self, key, value, default=False, **kwargs):
        super(Variant, self).__init__(**kwargs)
        self.key = key
        self.value = value
        self.default = default

class NamedArgument(SyntaxNode):
    def __init__(self, name, val, **kwargs):
        super(NamedArgument, self).__init__(**kwargs)
        self.name = name
        self.val = val

class Identifier(SyntaxNode):
    def __init__(self, name, **kwargs):
        super(Identifier, self).__init__(**kwargs)
        self.name = name

class VariantName(Identifier):
    def __init__(self, name, **kwargs):
        super(VariantName, self).__init__(name, **kwargs)


class BaseComment(Entry):
    def __init__(self, content=None, **kwargs):
        super(BaseComment, self).__init__(**kwargs)
        self.content = content


class Comment(BaseComment):
    def __init__(self, content=None, **kwargs):
        super(Comment, self).__init__(content, **kwargs)


class GroupComment(BaseComment):
    def __init__(self, content=None, **kwargs):
        super(GroupComment, self).__init__(content, **kwargs)


class ResourceComment(BaseComment):
    def __init__(self, content=None, **kwargs):
        super(ResourceComment, self).__init__(content, **kwargs)


class Function(Identifier):
    def __init__(self, name, **kwargs):
        super(Function, self).__init__(name, **kwargs)

class Junk(Entry):
    def __init__(self, content=None, **kwargs):
        super(Junk, self).__init__(**kwargs)
        self.content = content


class Span(BaseNode):
    def __init__(self, start, end, **kwargs):
        super(Span, self).__init__(**kwargs)
        self.start = start
        self.end = end


class Annotation(SyntaxNode):
    def __init__(self, code, args=None, message=None, **kwargs):
        super(Annotation, self).__init__(**kwargs)
        self.code = code
        self.args = args or []
        self.message = message



#!/usr/bin/env python

"""
test .ini parsing

ensure our .ini parser is doing what we want; to be deprecated for
python's standard ConfigParser when 2.7 is reality so OrderedDict
is the default:

http://docs.python.org/2/library/configparser.html
"""

from __future__ import absolute_import

import unittest
from manifestparser import read_ini
from StringIO import StringIO

import mozunit


class IniParserTest(unittest.TestCase):

    def parse_manifest(self, string):
        buf = StringIO()
        buf.write(string)
        buf.seek(0)
        return read_ini(buf)

    def test_inline_comments(self):
        result = self.parse_manifest("""
[test_felinicity.py]
kittens = true # This test requires kittens
cats = false#but not cats
""")[0][1]

        # make sure inline comments get stripped out, but comments without a space in front don't
        self.assertEqual(result['kittens'], 'true')
        self.assertEqual(result['cats'], "false#but not cats")

    def test_line_continuation(self):
        result = self.parse_manifest("""
[test_caninicity.py]
breeds =
  sheppard
  retriever
  terrier

[test_cats_and_dogs.py]
  cats=yep
  dogs=
    yep
      yep
birds=nope
  fish=nope
""")
        self.assertEqual(result[0][1]['breeds'].split(), ['sheppard', 'retriever', 'terrier'])
        self.assertEqual(result[1][1]['cats'], 'yep')
        self.assertEqual(result[1][1]['dogs'].split(), ['yep', 'yep'])
        self.assertEqual(result[1][1]['birds'].split(), ['nope', 'fish=nope'])


if __name__ == '__main__':
    mozunit.main()



from __future__ import absolute_import, print_function, unicode_literals

import math
from .shared import string, to_str, fromNow, JSONTemplateError


class BuiltinError(JSONTemplateError):
    pass


def build(context):
    builtins = {}

    def builtin(name, variadic=None, argument_tests=None, minArgs=None):
        def wrap(fn):
            def bad(reason=None):
                raise BuiltinError(
                    (reason or 'invalid arguments to builtin: {}').format(name))
            if variadic:
                def invoke(*args):
                    if minArgs:
                        if len(args) < minArgs:
                            bad("too few arguments to {}")
                    for arg in args:
                        if not variadic(arg):
                            bad()
                    return fn(*args)

            elif argument_tests:
                def invoke(*args):
                    if len(args) != len(argument_tests):
                        bad()
                    for t, arg in zip(argument_tests, args):
                        if not t(arg):
                            bad()
                    return fn(*args)

            else:
                def invoke(*args):
                    return fn(*args)

            builtins[name] = invoke
            return fn
        return wrap

    def is_number(v):
        return isinstance(v, (int, float)) and not isinstance(v, bool)

    def is_string(v):
        return isinstance(v, string)

    def is_string_or_array(v):
        return isinstance(v, (string, list))

    def anything_except_array(v):
        return isinstance(v, (string, int, float, bool)) or v is None

    def anything(v):
        return isinstance(v, (string, int, float, list, dict)) or v is None or callable(v)

    # ---

    builtin('min', variadic=is_number, minArgs=1)(min)
    builtin('max', variadic=is_number, minArgs=1)(max)
    builtin('sqrt', argument_tests=[is_number])(math.sqrt)
    builtin('abs', argument_tests=[is_number])(abs)

    @builtin('ceil', argument_tests=[is_number])
    def ceil(v):
        return int(math.ceil(v))

    @builtin('floor', argument_tests=[is_number])
    def floor(v):
        return int(math.floor(v))

    @builtin('lowercase', argument_tests=[is_string])
    def lowercase(v):
        return v.lower()

    @builtin('uppercase', argument_tests=[is_string])
    def lowercase(v):
        return v.upper()

    builtin('len', argument_tests=[is_string_or_array])(len)
    builtin('str', argument_tests=[anything_except_array])(to_str)

    @builtin('strip', argument_tests=[is_string])
    def strip(s):
        return s.strip()

    @builtin('rstrip', argument_tests=[is_string])
    def rstrip(s):
        return s.rstrip()

    @builtin('lstrip', argument_tests=[is_string])
    def lstrip(s):
        return s.lstrip()

    @builtin('fromNow', variadic=is_string, minArgs=1)
    def fromNow_builtin(offset, reference=None):
        return fromNow(offset, reference or context.get('now'))

    @builtin('typeof', argument_tests=[anything])
    def typeof(v):
        if isinstance(v, bool):
            return 'boolean'
        elif isinstance(v, string):
            return 'string'
        elif isinstance(v, (int, float)):
            return 'number'
        elif isinstance(v, list):
            return 'array'
        elif isinstance(v, dict):
            return 'object'
        elif v is None:
            return None
        elif callable(v):
            return 'function'

    return builtins



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import unicode_literals

import os
import shutil
import tempfile
import unittest

from copy import deepcopy

from mozunit import main

from mozbuild.test.backend.common import BackendTester

import mozpack.path as mozpath

from mozbuild.gn_processor import (
    GnMozbuildWriterBackend,
    find_common_attrs,
)
from mozbuild.backend.recursivemake import RecursiveMakeBackend

from mozbuild.frontend.data import (
    ComputedFlags,
    StaticLibrary,
    Sources,
    UnifiedSources,
)

class TestGnMozbuildWriter(BackendTester):

    def setUp(self):
        self._backup_dir = tempfile.mkdtemp(prefix='mozbuild')
        self._gn_dir = os.path.join(os.path.dirname(__file__),
                                    'data', 'gn-processor', 'trunk')
        shutil.copytree(os.path.join(self._gn_dir),
                        os.path.join(self._backup_dir, 'trunk'))
        super(TestGnMozbuildWriter, self).setUp()

    def tearDown(self):
        shutil.rmtree(self._gn_dir)
        shutil.copytree(os.path.join(self._backup_dir, 'trunk'),
                        self._gn_dir)
        shutil.rmtree(self._backup_dir)
        super(TestGnMozbuildWriter, self).tearDown()

    def test_gn_processor(self):
        # Generate our moz.build files.
        env = self._get_environment('gn-processor')
        self._consume('gn-processor', GnMozbuildWriterBackend, env=env)

        # Read the resulting tree.
        _, objs = self._emit('gn-processor', env=env)

        # The config for this srcdir sets OS_TARGET = Darwin, for which our gn config
        # generates two trivial libraries that link in to the main webrtc library,
        # one source file, one unified source file, and one directory with some
        # local defines and includes set. The following asserts this is what we see when
        # we read the generated moz.build files.
        objs = list(objs)
        expected_srcdirs = set([
            mozpath.join(env.topsrcdir, p)
            for p in ['trunk',
                      'trunk/webrtc/webrtc_gn',
                      'trunk/webrtc/base/mac_base_gn']
        ] + [env.topsrcdir])
        actual_srcdirs = set([
            o.srcdir for o in objs
        ])
        self.assertEqual(expected_srcdirs, actual_srcdirs)

        libs = {o.lib_name: o for o in objs if isinstance(o, StaticLibrary)}
        expected_libs = set(['libwebrtc.a', 'libwebrtc_gn.a', 'libmac_base_gn.a'])
        self.assertEqual(expected_libs, set(libs.keys()))
        # All libs are linked into 'libwebrtc.a'.
        self.assertEqual(set(libs['libwebrtc.a'].linked_libraries),
                         set(libs.values()) - set([libs['libwebrtc.a']]))

        [sources] = [o for o in objs if isinstance(o, Sources)]
        [unified_sources] = [o for o in objs if isinstance(o, UnifiedSources)]
        self.assertEqual(sources.files,
                         [mozpath.join(sources.topsrcdir, 'trunk', 'webrtc',
                                       'build', 'function.cc')])
        self.assertEqual(unified_sources.files,
                         [mozpath.join(unified_sources.topsrcdir, 'trunk', 'webrtc',
                                       'build', 'no_op_function.cc')])

        [flags_obj] = [o for o in objs if isinstance(o, ComputedFlags)
                       if 'DEFINES' in o.flags and o.flags['DEFINES']]
        self.assertEqual(flags_obj.srcdir, mozpath.join(flags_obj.topsrcdir,
                                                        'trunk', 'webrtc', 'webrtc_gn'))
        expected_defines = set([
            '-DUSE_NSS_CERTS=1',
            '-DCHROMIUM_BUILD',
            '-DUSE_X11=1',
            '-DNDEBUG',
            '-DNVALGRIND',
            '-DWEBRTC_MAC',
        ])
        self.assertEqual(set(flags_obj.flags['DEFINES']), expected_defines)

        expected_includes = set([
            '-I' + mozpath.join(flags_obj.topobjdir, 'ipc', 'ipdl', '_ipdlheaders'),
            '-I' + mozpath.join(flags_obj.topsrcdir, 'ipc', 'chromium', 'src'),
            '-I' + mozpath.join(flags_obj.topsrcdir, 'ipc', 'glue'),
            '-I' + mozpath.join(flags_obj.topsrcdir, 'trunk', 'webrtc', 'modules', 'include'),
        ])
        self.assertEqual(set(flags_obj.flags['LOCAL_INCLUDES']), expected_includes)
        self.assertEqual(flags_obj.flags['MOZBUILD_CXXFLAGS'], ['-fstack-protector'])


class TestGnMozbuildFactoring(unittest.TestCase):

    test_attrs = {
        'LOCAL_INCLUDES': [
            '!/ipc/ipdl/_ipdlheaders',
            '/ipc/chromium/src',
            '/ipc/glue',
        ],
        'DEFINES': {
            'UNICODE': True,
            'DEBUG': True,
        },
        'LIBRARY_NAME': 'Foo',
    }

    def test_single_attr_set(self):
        # If we only have one set of attributes they're all considered common.
        single_attrs = deepcopy(self.test_attrs)
        common_attrs = find_common_attrs([single_attrs])
        self.assertFalse(any(v for v in single_attrs.values()))
        self.assertEqual(self.test_attrs, common_attrs)

    def test_list_factor(self):
        list_attrs = deepcopy(self.test_attrs)
        input_attrs = deepcopy(self.test_attrs)
        list_attrs['LOCAL_INCLUDES'] += ['/local/include']
        list_attrs['LOCAL_INCLDUES'] = list(reversed(list_attrs['LOCAL_INCLUDES']))
        common_attrs = find_common_attrs([input_attrs, list_attrs])
        self.assertEqual(list_attrs['LOCAL_INCLUDES'], ['/local/include'])
        self.assertEqual(self.test_attrs, common_attrs)
        self.assertFalse(any(v for v in input_attrs.values()))

    def test_value_factor(self):
        input_attrs = deepcopy(self.test_attrs)
        value_attrs = deepcopy(self.test_attrs)
        value_attrs['LOCAL_INCLUDES'] = ['/local/include']
        value_attrs['DEFINES']['DEBUG'] = False
        value_attrs['DEFINES']['UNICODE'] = False
        common_attrs = find_common_attrs([value_attrs, input_attrs])
        self.assertEqual(common_attrs['LIBRARY_NAME'], 'Foo')
        self.assertNotIn('LIBRARY_NAME', value_attrs)
        self.assertNotIn('LIBRARY_NAME', input_attrs)
        del common_attrs['LIBRARY_NAME']
        self.assertFalse(any(v for v in common_attrs.values()))
        self.assertEqual(input_attrs['LOCAL_INCLUDES'], self.test_attrs['LOCAL_INCLUDES'])
        self.assertEqual(input_attrs['DEFINES'], self.test_attrs['DEFINES'])
        self.assertEqual(value_attrs['LOCAL_INCLUDES'],  ['/local/include'])
        self.assertEqual(value_attrs['DEFINES'],  {'DEBUG': False, 'UNICODE': False})

    def test_dictionary_factor(self):
        input_attrs = deepcopy(self.test_attrs)
        dict_subset_attrs = deepcopy(self.test_attrs)
        overlapping_dict_attrs = deepcopy(self.test_attrs)

        del dict_subset_attrs['DEFINES']['DEBUG']
        overlapping_dict_attrs['DEFINES']['MOZ_DEBUG'] = True

        common_attrs = find_common_attrs([input_attrs, dict_subset_attrs, overlapping_dict_attrs])
        for attr in ('LIBRARY_NAME', 'LOCAL_INCLUDES'):
            self.assertEqual(common_attrs[attr], self.test_attrs[attr])
            self.assertFalse(input_attrs.get(attr))
            self.assertFalse(dict_subset_attrs.get(attr))
            self.assertFalse(overlapping_dict_attrs.get(attr))
        self.assertEqual(common_attrs['DEFINES'], {'UNICODE': True})
        self.assertEqual(dict_subset_attrs['DEFINES'], {})
        self.assertEqual(overlapping_dict_attrs['DEFINES'], {
            'DEBUG': True,
            'MOZ_DEBUG': True,
        })

    def test_common_attrs(self):
        input_attrs = deepcopy(self.test_attrs)
        dict_subset_attrs = deepcopy(self.test_attrs)

        del dict_subset_attrs['DEFINES']
        common_attrs = find_common_attrs([input_attrs, dict_subset_attrs])
        self.assertNotIn('DEFINES', common_attrs)
        expected_common = deepcopy(self.test_attrs)
        del expected_common['DEFINES']
        self.assertEqual(expected_common, common_attrs)
        for attr_set in (input_attrs, dict_subset_attrs):
            self.assertNotIn('LIBRARY_NAME', attr_set)
            self.assertEqual(attr_set['LOCAL_INCLUDES'], [])


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
Unit-Tests for moznetwork
"""

from __future__ import absolute_import

import mock
import mozinfo
import moznetwork
import re
import subprocess
from distutils.spawn import find_executable

import mozunit
import pytest


@pytest.fixture(scope='session')
def ip_addresses():
    """List of IP addresses associated with the host."""

    # Regex to match IPv4 addresses.
    # 0-255.0-255.0-255.0-255, note order is important here.
    regexip = re.compile("((25[0-5]|2[0-4]\d|1\d\d|[1-9]\d|\d)\.){3}"
                         "(25[0-5]|2[0-4]\d|1\d\d|[1-9]\d|\d)")

    commands = (
        ['ip', 'addr', 'show'],
        ['ifconfig'],
        ['ipconfig'],
        # Explicitly search '/sbin' because it doesn't always appear
        # to be on the $PATH of all systems
        ['/sbin/ip', 'addr', 'show'],
        ['/sbin/ifconfig'],
    )

    cmd = None
    for command in commands:
        if find_executable(command[0]):
            cmd = command
            break
    else:
        raise OSError("No program for detecting ip address found! Ensure one of 'ip', "
                      "'ifconfig' or 'ipconfig' exists on your $PATH.")

    ps = subprocess.Popen(cmd, stdout=subprocess.PIPE)
    standardoutput, standarderror = ps.communicate()

    # Generate a list of IPs by parsing the output of ip/ifconfig
    return [x.group() for x in re.finditer(regexip, standardoutput)]


def test_get_ip(ip_addresses):
    """ Attempt to test the IP address returned by
    moznetwork.get_ip() is valid """
    assert moznetwork.get_ip() in ip_addresses


def test_get_ip_using_get_interface(ip_addresses):
    """ Test that the control flow path for get_ip() using
    _get_interface_list() is works """

    if mozinfo.isLinux or mozinfo.isMac:

        with mock.patch('socket.gethostbyname') as byname:
            # Force socket.gethostbyname to return None
            byname.return_value = None
            assert moznetwork.get_ip() in ip_addresses


if __name__ == '__main__':
    mozunit.main()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import errno
import os
import tempfile
import tarfile
import shutil
import mozpack.path as mozpath
from mozpack.dmg import create_dmg
from application_ini import get_application_ini_value

def repackage_dmg(infile, output):

    if not tarfile.is_tarfile(infile):
        raise Exception("Input file %s is not a valid tarfile." % infile)

    tmpdir = tempfile.mkdtemp()
    try:
        with tarfile.open(infile) as tar:
            tar.extractall(path=tmpdir)

        # Remove the /Applications symlink. If we don't, an rsync command in
        # create_dmg() will break, and create_dmg() re-creates the symlink anyway.
        try:
            os.remove(mozpath.join(tmpdir, ' '))
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise

        volume_name = get_application_ini_value(tmpdir, 'App', 'CodeName',
                                                fallback='Name')

        # The extra_files argument is empty [] because they are already a part
        # of the original dmg produced by the build, and they remain in the
        # tarball generated by the signing task.
        create_dmg(tmpdir, output, volume_name, [])

    finally:
        shutil.rmtree(tmpdir)



import pytest

from cbor2.types import CBORTag, CBORSimpleValue


def test_tag_repr():
    assert repr(CBORTag(600, 'blah')) == "CBORTag(600, 'blah')"


def test_tag_equals():
    tag1 = CBORTag(500, ['foo'])
    tag2 = CBORTag(500, ['foo'])
    tag3 = CBORTag(500, ['bar'])
    assert tag1 == tag2
    assert not tag1 == tag3
    assert not tag1 == 500


def test_simple_value_repr():
    assert repr(CBORSimpleValue(1)) == "CBORSimpleValue(1)"


def test_simple_value_equals():
    tag1 = CBORSimpleValue(1)
    tag2 = CBORSimpleValue(1)
    tag3 = CBORSimpleValue(21)
    assert tag1 == tag2
    assert tag1 == 1
    assert not tag1 == tag3
    assert not tag1 == 21
    assert not tag2 == "21"


def test_simple_value_too_big():
    exc = pytest.raises(TypeError, CBORSimpleValue, 256)
    assert str(exc.value) == 'simple value too big'



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import absolute_import, print_function, unicode_literals

import unittest
import mozunit
from taskgraph.util.templates import (
    merge_to,
    merge
)


class MergeTest(unittest.TestCase):

    def test_merge_to_dicts(self):
        source = {'a': 1, 'b': 2}
        dest = {'b': '20', 'c': 30}
        expected = {
            'a': 1,   # source only
            'b': 2,   # source overrides dest
            'c': 30,  # dest only
        }
        self.assertEqual(merge_to(source, dest), expected)
        self.assertEqual(dest, expected)

    def test_merge_to_lists(self):
        source = {'x': [3, 4]}
        dest = {'x': [1, 2]}
        expected = {'x': [1, 2, 3, 4]}  # dest first
        self.assertEqual(merge_to(source, dest), expected)
        self.assertEqual(dest, expected)

    def test_merge_diff_types(self):
        source = {'x': [1, 2]}
        dest = {'x': 'abc'}
        expected = {'x': [1, 2]}  # source wins
        self.assertEqual(merge_to(source, dest), expected)
        self.assertEqual(dest, expected)

    def test_merge(self):
        first = {'a': 1, 'b': 2, 'd': 11}
        second = {'b': 20, 'c': 30}
        third = {'c': 300, 'd': 400}
        expected = {
            'a': 1,
            'b': 20,
            'c': 300,
            'd': 400,
        }
        self.assertEqual(merge(first, second, third), expected)

        # inputs haven't changed..
        self.assertEqual(first, {'a': 1, 'b': 2, 'd': 11})
        self.assertEqual(second, {'b': 20, 'c': 30})
        self.assertEqual(third, {'c': 300, 'd': 400})


if __name__ == '__main__':
    mozunit.main()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import unittest

from compare_locales import util


class ParseLocalesTest(unittest.TestCase):
    def test_empty(self):
        self.assertEquals(util.parseLocales(''), [])

    def test_all(self):
        self.assertEquals(util.parseLocales('''af
de'''), ['af', 'de'])

    def test_shipped(self):
        self.assertEquals(util.parseLocales('''af
ja win mac
de'''), ['af', 'de', 'ja'])

    def test_sparse(self):
        self.assertEquals(util.parseLocales('''
af

de

'''), ['af', 'de'])



from __future__ import unicode_literals
import re
from .ftlstream import FTLParserStream
from . import ast
from .errors import ParseError


def with_span(fn):
    def decorated(self, ps, *args):
        if not self.with_spans:
            return fn(self, ps, *args)

        start = ps.get_index()
        node = fn(self, ps, *args)

        # Don't re-add the span if the node already has it.  This may happen
        # when one decorated function calls another decorated function.
        if node.span is not None:
            return node

        # Spans of Messages should include the attached Comment.
        if isinstance(node, ast.Message):
            if node.comment is not None:
                start = node.comment.span.start

        end = ps.get_index()
        node.add_span(start, end)
        return node

    return decorated


class FluentParser(object):
    def __init__(self, with_spans=True):
        self.with_spans = with_spans

    def parse(self, source):
        ps = FTLParserStream(source)
        ps.skip_blank_lines()

        entries = []

        while ps.current():
            entry = self.get_entry_or_junk(ps)

            if entry is None:
                continue

            if isinstance(entry, ast.Comment) \
               and ps.last_comment_zero_four_syntax \
               and len(entries) == 0:
                comment = ast.ResourceComment(entry.content)
                comment.span = entry.span
                entries.append(comment)
            else:
                entries.append(entry)

            ps.last_comment_zero_four_syntax = False
            ps.skip_blank_lines()

        res = ast.Resource(entries)

        if self.with_spans:
            res.add_span(0, ps.get_index())

        return res

    def parse_entry(self, source):
        ps = FTLParserStream(source)
        ps.skip_blank_lines()
        return self.get_entry_or_junk(ps)

    def get_entry_or_junk(self, ps):
        entry_start_pos = ps.get_index()

        try:
            return self.get_entry(ps)
        except ParseError as err:
            error_index = ps.get_index()
            ps.skip_to_next_entry_start()
            next_entry_start = ps.get_index()

            # Create a Junk instance
            slice = ps.get_slice(entry_start_pos, next_entry_start)
            junk = ast.Junk(slice)
            if self.with_spans:
                junk.add_span(entry_start_pos, next_entry_start)
            annot = ast.Annotation(err.code, err.args, err.message)
            annot.add_span(error_index, error_index)
            junk.add_annotation(annot)
            return junk

    def get_entry(self, ps):
        comment = None

        if ps.current_is('/') or ps.current_is('#'):
            comment = self.get_comment(ps)

            # The Comment content doesn't include the trailing newline. Consume
            # this newline to be ready for the next entry. None stands for EOF.
            ps.expect_char('\n' if ps.current() else None)

        if ps.current_is('['):
            group_comment = self.get_group_comment_from_section(ps, comment)
            if comment and self.with_spans:
                # The Group Comment should start where the section comment
                # starts.
                group_comment.span.start = comment.span.start
            return group_comment

        if ps.is_entry_id_start() \
           and (comment is None or isinstance(comment, ast.Comment)):
            return self.get_message(ps, comment)

        if comment:
            return comment

        raise ParseError('E0002')

    @with_span
    def get_zero_four_style_comment(self, ps):
        ps.expect_char('/')
        ps.expect_char('/')
        ps.take_char_if(' ')

        content = ''

        while True:
            ch = ps.take_char(lambda x: x != '\n')
            while ch:
                content += ch
                ch = ps.take_char(lambda x: x != '\n')

            if ps.is_peek_next_line_zero_four_style_comment():
                content += '\n'
                ps.next()
                ps.expect_char('/')
                ps.expect_char('/')
                ps.take_char_if(' ')
            else:
                break

        comment = ast.Comment(content)
        ps.last_comment_zero_four_syntax = True
        return comment

    @with_span
    def get_comment(self, ps):
        if ps.current_is('/'):
            return self.get_zero_four_style_comment(ps)

        # 0 - comment
        # 1 - group comment
        # 2 - resource comment
        level = -1
        content = ''

        while True:
            i = -1
            while ps.current_is('#') and (i < (2 if level == -1 else level)):
                ps.next()
                i += 1

            if level == -1:
                level = i

            if not ps.current_is('\n'):
                ps.expect_char(' ')
                ch = ps.take_char(lambda x: x != '\n')
                while ch:
                    content += ch
                    ch = ps.take_char(lambda x: x != '\n')

            if ps.is_peek_next_line_comment(level):
                content += '\n'
                ps.next()
            else:
                break

        if level == 0:
            return ast.Comment(content)
        elif level == 1:
            return ast.GroupComment(content)
        elif level == 2:
            return ast.ResourceComment(content)

    @with_span
    def get_group_comment_from_section(self, ps, comment):
        ps.expect_char('[')
        ps.expect_char('[')

        ps.skip_inline_ws()

        self.get_variant_name(ps)

        ps.skip_inline_ws()

        ps.expect_char(']')
        ps.expect_char(']')

        if comment:
            return ast.GroupComment(comment.content)

        # A Section without a comment is like an empty Group Comment.
        # Semantically it ends the previous group and starts a new one.
        return ast.GroupComment('')

    @with_span
    def get_message(self, ps, comment):
        id = self.get_entry_identifier(ps)

        ps.skip_inline_ws()

        pattern = None
        attrs = None

        # XXX Syntax 0.4 compat
        if ps.current_is('='):
            ps.next()

            if ps.is_peek_pattern_start():
                ps.skip_indent()
                pattern = self.get_pattern(ps)

        if id.name.startswith('-') and pattern is None:
            raise ParseError('E0006', id.name)

        if ps.is_peek_next_line_attribute_start():
            attrs = self.get_attributes(ps)

        if id.name.startswith('-'):
            return ast.Term(id, pattern, attrs, comment)

        if pattern is None and attrs is None:
            raise ParseError('E0005', id.name)

        return ast.Message(id, pattern, attrs, comment)

    @with_span
    def get_attribute(self, ps):
        ps.expect_char('.')

        key = self.get_identifier(ps)

        ps.skip_inline_ws()
        ps.expect_char('=')

        if ps.is_peek_pattern_start():
            ps.skip_indent()
            value = self.get_pattern(ps)
            return ast.Attribute(key, value)

        raise ParseError('E0012')

    def get_attributes(self, ps):
        attrs = []

        while True:
            ps.expect_indent()
            attr = self.get_attribute(ps)
            attrs.append(attr)

            if not ps.is_peek_next_line_attribute_start():
                break
        return attrs

    def get_entry_identifier(self, ps):
        return self.get_identifier(ps, True)

    @with_span
    def get_identifier(self, ps, allow_term=False):
        name = ''

        name += ps.take_id_start(allow_term)

        ch = ps.take_id_char()
        while ch:
            name += ch
            ch = ps.take_id_char()

        return ast.Identifier(name)

    def get_variant_key(self, ps):
        ch = ps.current()

        if ch is None:
            raise ParseError('E0013')

        cc = ord(ch)
        if ((cc >= 48 and cc <= 57) or cc == 45):  # 0-9, -
            return self.get_number(ps)

        return self.get_variant_name(ps)

    @with_span
    def get_variant(self, ps, has_default):
        default_index = False

        if ps.current_is('*'):
            if has_default:
                raise ParseError('E0015')
            ps.next()
            default_index = True

        ps.expect_char('[')

        key = self.get_variant_key(ps)

        ps.expect_char(']')

        if ps.is_peek_pattern_start():
            ps.skip_indent()
            value = self.get_pattern(ps)
            return ast.Variant(key, value, default_index)

        raise ParseError('E0012')

    def get_variants(self, ps):
        variants = []
        has_default = False

        while True:
            ps.expect_indent()
            variant = self.get_variant(ps, has_default)

            if variant.default:
                has_default = True

            variants.append(variant)

            if not ps.is_peek_next_line_variant_start():
                break

        if not has_default:
            raise ParseError('E0010')

        return variants

    @with_span
    def get_variant_name(self, ps):
        name = ''

        name += ps.take_id_start(False)

        while True:
            ch = ps.take_variant_name_char()
            if ch:
                name += ch
            else:
                break

        return ast.VariantName(name.rstrip())

    def get_digits(self, ps):
        num = ''

        ch = ps.take_digit()
        while ch:
            num += ch
            ch = ps.take_digit()

        if len(num) == 0:
            raise ParseError('E0004', '0-9')

        return num

    @with_span
    def get_number(self, ps):
        num = ''

        if ps.current_is('-'):
            num += '-'
            ps.next()

        num += self.get_digits(ps)

        if ps.current_is('.'):
            num += '.'
            ps.next()
            num += self.get_digits(ps)

        return ast.NumberExpression(num)

    @with_span
    def get_pattern(self, ps):
        elements = []
        ps.skip_inline_ws()

        while ps.current():
            ch = ps.current()

            # The end condition for get_pattern's while loop is a newline
            # which is not followed by a valid pattern continuation.
            if ch == '\n' and not ps.is_peek_next_line_pattern_start():
                break

            if ch == '{':
                element = self.get_placeable(ps)
            else:
                element = self.get_text_element(ps)

            elements.append(element)

        return ast.Pattern(elements)

    @with_span
    def get_text_element(self, ps):
        buf = ''

        while ps.current():
            ch = ps.current()

            if ch == '{':
                return ast.TextElement(buf)

            if ch == '\n':
                if not ps.is_peek_next_line_pattern_start():
                    return ast.TextElement(buf)

                ps.next()
                ps.skip_inline_ws()

                # Add the new line to the buffer
                buf += ch
                continue

            if ch == '\\':
                ch2 = ps.next()

                if ch2 == '{' or ch2 == '"':
                    buf += ch2
                else:
                    buf += ch + ch2

            else:
                buf += ch

            ps.next()

        return ast.TextElement(buf)

    @with_span
    def get_placeable(self, ps):
        ps.expect_char('{')
        expression = self.get_expression(ps)
        ps.expect_char('}')
        return ast.Placeable(expression)

    @with_span
    def get_expression(self, ps):

        if ps.is_peek_next_line_variant_start():
            variants = self.get_variants(ps)

            ps.expect_indent()

            return ast.SelectExpression(None, variants)

        ps.skip_inline_ws()

        selector = self.get_selector_expression(ps)

        ps.skip_inline_ws()

        if ps.current_is('-'):
            ps.peek()

            if not ps.current_peek_is('>'):
                ps.reset_peek()
                return selector

            if isinstance(selector, ast.MessageReference):
                raise ParseError('E0016')

            if isinstance(selector, ast.AttributeExpression) and \
               not selector.id.name.startswith('-'):
                raise ParseError('E0018')

            if isinstance(selector, ast.VariantExpression):
                raise ParseError('E0017')

            ps.next()
            ps.next()

            ps.skip_inline_ws()

            variants = self.get_variants(ps)

            if len(variants) == 0:
                raise ParseError('E0011')

            ps.expect_indent()

            return ast.SelectExpression(selector, variants)
        elif isinstance(selector, ast.AttributeExpression) and \
                selector.id.name.startswith('-'):
            raise ParseError('E0019')

        return selector

    @with_span
    def get_selector_expression(self, ps):
        literal = self.get_literal(ps)

        if not isinstance(literal, ast.MessageReference):
            return literal

        ch = ps.current()

        if (ch == '.'):
            ps.next()
            attr = self.get_identifier(ps)
            return ast.AttributeExpression(literal.id, attr)

        if (ch == '['):
            ps.next()
            key = self.get_variant_key(ps)
            ps.expect_char(']')
            return ast.VariantExpression(literal.id, key)

        if (ch == '('):
            ps.next()

            args = self.get_call_args(ps)

            ps.expect_char(')')

            if not re.match('^[A-Z][A-Z_?-]*$', literal.id.name):
                raise ParseError('E0008')

            return ast.CallExpression(
                ast.Function(literal.id.name),
                args
            )

        return literal

    @with_span
    def get_call_arg(self, ps):
        exp = self.get_selector_expression(ps)

        ps.skip_inline_ws()

        if not ps.current_is(':'):
            return exp

        if not isinstance(exp, ast.MessageReference):
            raise ParseError('E0009')

        ps.next()
        ps.skip_inline_ws()

        val = self.get_arg_val(ps)

        return ast.NamedArgument(exp.id, val)

    def get_call_args(self, ps):
        args = []

        ps.skip_inline_ws()

        while True:
            if ps.current_is(')'):
                break

            arg = self.get_call_arg(ps)
            args.append(arg)

            ps.skip_inline_ws()

            if ps.current_is(','):
                ps.next()
                ps.skip_inline_ws()
                continue
            else:
                break

        return args

    def get_arg_val(self, ps):
        if ps.is_number_start():
            return self.get_number(ps)
        elif ps.current_is('"'):
            return self.get_string(ps)
        raise ParseError('E0012')

    @with_span
    def get_string(self, ps):
        val = ''

        ps.expect_char('"')

        ch = ps.take_char(lambda x: x != '"' and x != '\n')
        while ch:
            val += ch
            ch = ps.take_char(lambda x: x != '"' and x != '\n')

        if ps.current_is('\n'):
            raise ParseError('E0020')

        ps.next()

        return ast.StringExpression(val)

    @with_span
    def get_literal(self, ps):
        ch = ps.current()

        if ch is None:
            raise ParseError('E0014')

        if ch == '$':
            ps.next()
            name = self.get_identifier(ps)
            return ast.ExternalArgument(name)
        elif ps.is_entry_id_start():
            name = self.get_entry_identifier(ps)
            return ast.MessageReference(name)
        elif ps.is_number_start():
            return self.get_number(ps)
        elif ch == '"':
            return self.get_string(ps)

        raise ParseError('E0014')



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import unicode_literals

import cPickle as pickle
import json
import os
import unittest

from mozpack.copier import FileRegistry
from mozpack.manifests import InstallManifest
from mozunit import main

from mozbuild.backend.fastermake import FasterMakeBackend

from mozbuild.test.backend.common import BackendTester

import mozpack.path as mozpath


class TestFasterMakeBackend(BackendTester):
    def test_basic(self):
        """Ensure the FasterMakeBackend works without error."""
        env = self._consume('stub0', FasterMakeBackend)
        self.assertTrue(os.path.exists(mozpath.join(env.topobjdir,
            'backend.FasterMakeBackend')))
        self.assertTrue(os.path.exists(mozpath.join(env.topobjdir,
            'backend.FasterMakeBackend.in')))

    def test_final_target_files_wildcard(self):
        """Ensure that wildcards in FINAL_TARGET_FILES work properly."""
        env = self._consume('final-target-files-wildcard', FasterMakeBackend)
        m = InstallManifest(path=mozpath.join(env.topobjdir,
            'faster', 'install_dist_bin'))
        self.assertEqual(len(m), 1)
        reg = FileRegistry()
        m.populate_registry(reg)
        expected = [('foo/bar.xyz', 'bar.xyz'), ('foo/foo.xyz', 'foo.xyz')]
        actual = [(path, mozpath.relpath(f.path, env.topsrcdir)) for (path, f) in reg]
        self.assertEqual(expected, actual)


if __name__ == '__main__':
    main()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

'Mozilla l10n compare locales tool'

import codecs
import os
import shutil
import re
from collections import defaultdict

from json import dumps

from compare_locales import parser
from compare_locales import paths, mozpath
from compare_locales.checks import getChecker


class Tree(object):
    def __init__(self, valuetype):
        self.branches = dict()
        self.valuetype = valuetype
        self.value = None

    def __getitem__(self, leaf):
        parts = []
        if isinstance(leaf, paths.File):
            parts = [] if not leaf.locale else [leaf.locale]
            if leaf.module:
                parts += leaf.module.split('/')
            parts += leaf.file.split('/')
        else:
            parts = leaf.split('/')
        return self.__get(parts)

    def __get(self, parts):
        common = None
        old = None
        new = tuple(parts)
        t = self
        for k, v in self.branches.iteritems():
            for i, part in enumerate(zip(k, parts)):
                if part[0] != part[1]:
                    i -= 1
                    break
            if i < 0:
                continue
            i += 1
            common = tuple(k[:i])
            old = tuple(k[i:])
            new = tuple(parts[i:])
            break
        if old:
            self.branches.pop(k)
            t = Tree(self.valuetype)
            t.branches[old] = v
            self.branches[common] = t
        elif common:
            t = self.branches[common]
        if new:
            if common:
                return t.__get(new)
            t2 = t
            t = Tree(self.valuetype)
            t2.branches[new] = t
        if t.value is None:
            t.value = t.valuetype()
        return t.value

    indent = '  '

    def getContent(self, depth=0):
        '''
        Returns iterator of (depth, flag, key_or_value) tuples.
        If flag is 'value', key_or_value is a value object, otherwise
        (flag is 'key') it's a key string.
        '''
        keys = self.branches.keys()
        keys.sort()
        if self.value is not None:
            yield (depth, 'value', self.value)
        for key in keys:
            yield (depth, 'key', key)
            for child in self.branches[key].getContent(depth + 1):
                yield child

    def toJSON(self):
        '''
        Returns this Tree as a JSON-able tree of hashes.
        Only the values need to take care that they're JSON-able.
        '''
        if self.value is not None:
            return self.value
        return dict(('/'.join(key), self.branches[key].toJSON())
                    for key in self.branches.keys())

    def getStrRows(self):
        def tostr(t):
            if t[1] == 'key':
                return self.indent * t[0] + '/'.join(t[2])
            return self.indent * (t[0] + 1) + str(t[2])

        return map(tostr, self.getContent())

    def __str__(self):
        return '\n'.join(self.getStrRows())


class AddRemove(object):
    def __init__(self):
        self.left = self.right = None

    def set_left(self, left):
        if not isinstance(left, list):
            left = list(l for l in left)
        self.left = left

    def set_right(self, right):
        if not isinstance(right, list):
            right = list(l for l in right)
        self.right = right

    def __iter__(self):
        # order_map stores index in left and then index in right
        order_map = dict((item, (i, -1)) for i, item in enumerate(self.left))
        left_items = set(order_map)
        # as we go through the right side, keep track of which left
        # item we had in right last, and for items not in left,
        # set the sortmap to (left_offset, right_index)
        left_offset = -1
        right_items = set()
        for i, item in enumerate(self.right):
            right_items.add(item)
            if item in order_map:
                left_offset = order_map[item][0]
            else:
                order_map[item] = (left_offset, i)
        for item in sorted(order_map, key=lambda item: order_map[item]):
            if item in left_items and item in right_items:
                yield ('equal', item)
            elif item in left_items:
                yield ('delete', item)
            else:
                yield ('add', item)


class Observer(object):

    def __init__(self, quiet=0, filter=None, file_stats=False):
        '''Create Observer
        For quiet=1, skip per-entity missing and obsolete strings,
        for quiet=2, skip missing and obsolete files. For quiet=3,
        skip warnings and errors.
        '''
        self.summary = defaultdict(lambda: defaultdict(int))
        self.details = Tree(list)
        self.quiet = quiet
        self.filter = filter
        self.file_stats = None
        if file_stats:
            self.file_stats = defaultdict(lambda: defaultdict(dict))

    # support pickling
    def __getstate__(self):
        state = dict(summary=self._dictify(self.summary), details=self.details)
        if self.file_stats is not None:
            state['file_stats'] = self._dictify(self.file_stats)
        return state

    def __setstate__(self, state):
        self.summary = defaultdict(lambda: defaultdict(int))
        if 'summary' in state:
            for loc, stats in state['summary'].iteritems():
                self.summary[loc].update(stats)
        self.file_stats = None
        if 'file_stats' in state:
            self.file_stats = defaultdict(lambda: defaultdict(dict))
            for k, d in state['file_stats'].iteritems():
                self.file_stats[k].update(d)
        self.details = state['details']
        self.filter = None

    def _dictify(self, d):
        plaindict = {}
        for k, v in d.iteritems():
            plaindict[k] = dict(v)
        return plaindict

    def toJSON(self):
        # Don't export file stats, even if we collected them.
        # Those are not part of the data we use toJSON for.
        return {
            'summary': self._dictify(self.summary),
            'details': self.details.toJSON()
        }

    def updateStats(self, file, stats):
        # in multi-project scenarios, this file might not be ours,
        # check that.
        # Pass in a dummy entity key '' to avoid getting in to
        # generic file filters. If we have stats for those,
        # we want to aggregate the counts
        if (self.filter is not None and
                self.filter(file, entity='') == 'ignore'):
            return
        for category, value in stats.iteritems():
            self.summary[file.locale][category] += value
        if self.file_stats is None:
            return
        if 'missingInFiles' in stats:
            # keep track of how many strings are in a missing file
            # we got the {'missingFile': 'error'} from the notify pass
            self.details[file].append({'count': stats['missingInFiles']})
            # missingInFiles should just be "missing" in file stats
            self.file_stats[file.locale][file.localpath]['missing'] = \
                stats['missingInFiles']
            return  # there are no other stats for missing files
        self.file_stats[file.locale][file.localpath].update(stats)

    def notify(self, category, file, data):
        rv = 'error'
        if category in ['missingFile', 'obsoleteFile']:
            if self.filter is not None:
                rv = self.filter(file)
            if rv != "ignore" and self.quiet < 2:
                self.details[file].append({category: rv})
            return rv
        if category in ['missingEntity', 'obsoleteEntity']:
            if self.filter is not None:
                rv = self.filter(file, data)
            if rv == "ignore":
                return rv
            if self.quiet < 1:
                self.details[file].append({category: data})
            return rv
        if category in ('error', 'warning') and self.quiet < 3:
            self.details[file].append({category: data})
            self.summary[file.locale][category + 's'] += 1
        return rv

    def toExhibit(self):
        items = []
        for locale in sorted(self.summary.iterkeys()):
            summary = self.summary[locale]
            if locale is not None:
                item = {'id': 'xxx/' + locale,
                        'label': locale,
                        'locale': locale}
            else:
                item = {'id': 'xxx',
                        'label': 'xxx',
                        'locale': 'xxx'}
            item['type'] = 'Build'
            total = sum([summary[k]
                         for k in ('changed', 'unchanged', 'report', 'missing',
                                   'missingInFiles')
                         if k in summary])
            total_w = sum([summary[k]
                           for k in ('changed_w', 'unchanged_w', 'missing_w')
                           if k in summary])
            rate = (('changed' in summary and summary['changed'] * 100) or
                    0) / total
            item.update((k, summary.get(k, 0))
                        for k in ('changed', 'unchanged'))
            item.update((k, summary[k])
                        for k in ('report', 'errors', 'warnings')
                        if k in summary)
            item['missing'] = summary.get('missing', 0) + \
                summary.get('missingInFiles', 0)
            item['completion'] = rate
            item['total'] = total
            item.update((k, summary.get(k, 0))
                        for k in ('changed_w', 'unchanged_w', 'missing_w'))
            item['total_w'] = total_w
            result = 'success'
            if item.get('warnings', 0):
                result = 'warning'
            if item.get('errors', 0) or item.get('missing', 0):
                result = 'failure'
            item['result'] = result
            items.append(item)
        data = {
            "properties": dict.fromkeys(
                ("completion", "errors", "warnings", "missing", "report",
                 "missing_w", "changed_w", "unchanged_w",
                 "unchanged", "changed", "obsolete"),
                {"valueType": "number"}),
            "types": {
                "Build": {"pluralLabel": "Builds"}
            }}
        data['items'] = items
        return dumps(data, indent=2)

    def serialize(self, type="text"):
        if type == "exhibit":
            return self.toExhibit()
        if type == "json":
            return dumps(self.toJSON())

        def tostr(t):
            if t[1] == 'key':
                return '  ' * t[0] + '/'.join(t[2])
            o = []
            indent = '  ' * (t[0] + 1)
            for item in t[2]:
                if 'error' in item:
                    o += [indent + 'ERROR: ' + item['error']]
                elif 'warning' in item:
                    o += [indent + 'WARNING: ' + item['warning']]
                elif 'missingEntity' in item:
                    o += [indent + '+' + item['missingEntity']]
                elif 'obsoleteEntity' in item:
                    o += [indent + '-' + item['obsoleteEntity']]
                elif 'missingFile' in item:
                    o.append(indent + '// add and localize this file')
                elif 'obsoleteFile' in item:
                    o.append(indent + '// remove this file')
            return '\n'.join(o)

        out = []
        for locale, summary in sorted(self.summary.iteritems()):
            if locale is not None:
                out.append(locale + ':')
            out += [k + ': ' + str(v) for k, v in sorted(summary.iteritems())]
            total = sum([summary[k]
                         for k in ['changed', 'unchanged', 'report', 'missing',
                                   'missingInFiles']
                         if k in summary])
            rate = 0
            if total:
                rate = (('changed' in summary and summary['changed'] * 100) or
                        0) / total
            out.append('%d%% of entries changed' % rate)
        return '\n'.join(map(tostr, self.details.getContent()) + out)

    def __str__(self):
        return 'observer'


class ContentComparer:
    keyRE = re.compile('[kK]ey')
    nl = re.compile('\n', re.M)

    def __init__(self, observers, stat_observers=None):
        '''Create a ContentComparer.
        observer is usually a instance of Observer. The return values
        of the notify method are used to control the handling of missing
        entities.
        '''
        self.observers = observers
        if stat_observers is None:
            stat_observers = []
        self.stat_observers = stat_observers

    def create_merge_dir(self, merge_file):
        outdir = mozpath.dirname(merge_file)
        if not os.path.isdir(outdir):
            os.makedirs(outdir)

    def merge(self, ref_entities, ref_map, ref_file, l10n_file, merge_file,
              missing, skips, ctx, capabilities, encoding):

        if capabilities == parser.CAN_NONE:
            return

        if capabilities & parser.CAN_COPY and (skips or missing):
            self.create_merge_dir(merge_file)
            shutil.copyfile(ref_file.fullpath, merge_file)
            print "copied reference to " + merge_file
            return

        if not (capabilities & parser.CAN_SKIP):
            return

        # Start with None in case the merge file doesn't need to be created.
        f = None

        if skips:
            # skips come in ordered by key name, we need them in file order
            skips.sort(key=lambda s: s.span[0])

            # we need to skip a few erroneous blocks in the input, copy by hand
            self.create_merge_dir(merge_file)
            f = codecs.open(merge_file, 'wb', encoding)
            offset = 0
            for skip in skips:
                chunk = skip.span
                f.write(ctx.contents[offset:chunk[0]])
                offset = chunk[1]
            f.write(ctx.contents[offset:])

        if not (capabilities & parser.CAN_MERGE):
            return

        if skips or missing:
            if f is None:
                self.create_merge_dir(merge_file)
                shutil.copyfile(l10n_file.fullpath, merge_file)
                f = codecs.open(merge_file, 'ab', encoding)

            trailing = (['\n'] +
                        [ref_entities[ref_map[key]].all for key in missing] +
                        [ref_entities[ref_map[skip.key]].all for skip in skips
                         if not isinstance(skip, parser.Junk)])

            def ensureNewline(s):
                if not s.endswith('\n'):
                    return s + '\n'
                return s

            print "adding to " + merge_file
            f.write(''.join(map(ensureNewline, trailing)))

        if f is not None:
            f.close()

    def notify(self, category, file, data):
        """Check observer for the found data, and if it's
        not to ignore, notify stat_observers.
        """
        rvs = set(
            observer.notify(category, file, data)
            for observer in self.observers
            )
        if all(rv == 'ignore' for rv in rvs):
            return 'ignore'
        rvs.discard('ignore')
        for obs in self.stat_observers:
            # non-filtering stat_observers, ignore results
            obs.notify(category, file, data)
        if 'error' in rvs:
            return 'error'
        assert len(rvs) == 1
        return rvs.pop()

    def updateStats(self, file, stats):
        """Check observer for the found data, and if it's
        not to ignore, notify stat_observers.
        """
        for observer in self.observers + self.stat_observers:
            observer.updateStats(file, stats)

    def remove(self, obsolete):
        self.notify('obsoleteFile', obsolete, None)
        pass

    def compare(self, ref_file, l10n, merge_file, extra_tests=None):
        try:
            p = parser.getParser(ref_file.file)
        except UserWarning:
            # no comparison, XXX report?
            return
        try:
            p.readContents(ref_file.getContents())
        except Exception, e:
            self.notify('error', ref_file, str(e))
            return
        ref_entities, ref_map = p.parse()
        try:
            p.readContents(l10n.getContents())
            l10n_entities, l10n_map = p.parse()
            l10n_ctx = p.ctx
        except Exception, e:
            self.notify('error', l10n, str(e))
            return

        ar = AddRemove()
        ar.set_left(e.key for e in ref_entities)
        ar.set_right(e.key for e in l10n_entities)
        report = missing = obsolete = changed = unchanged = keys = 0
        missing_w = changed_w = unchanged_w = 0  # word stats
        missings = []
        skips = []
        checker = getChecker(l10n, extra_tests=extra_tests)
        if checker and checker.needs_reference:
            checker.set_reference(ref_entities)
        for msg in p.findDuplicates(ref_entities):
            self.notify('warning', l10n, msg)
        for msg in p.findDuplicates(l10n_entities):
            self.notify('error', l10n, msg)
        for action, entity_id in ar:
            if action == 'delete':
                # missing entity
                if isinstance(ref_entities[ref_map[entity_id]], parser.Junk):
                    self.notify('warning', l10n, 'Parser error in en-US')
                    continue
                _rv = self.notify('missingEntity', l10n, entity_id)
                if _rv == "ignore":
                    continue
                if _rv == "error":
                    # only add to missing entities for l10n-merge on error,
                    # not report
                    missings.append(entity_id)
                    missing += 1
                    refent = ref_entities[ref_map[entity_id]]
                    missing_w += refent.count_words()
                else:
                    # just report
                    report += 1
            elif action == 'add':
                # obsolete entity or junk
                if isinstance(l10n_entities[l10n_map[entity_id]],
                              parser.Junk):
                    junk = l10n_entities[l10n_map[entity_id]]
                    params = (junk.val,) + junk.position() + junk.position(-1)
                    self.notify('error', l10n,
                                'Unparsed content "%s" from line %d column %d'
                                ' to line %d column %d' % params)
                    if merge_file is not None:
                        skips.append(junk)
                elif self.notify('obsoleteEntity', l10n,
                                 entity_id) != 'ignore':
                    obsolete += 1
            else:
                # entity found in both ref and l10n, check for changed
                refent = ref_entities[ref_map[entity_id]]
                l10nent = l10n_entities[l10n_map[entity_id]]
                if self.keyRE.search(entity_id):
                    keys += 1
                else:
                    if refent.equals(l10nent):
                        self.doUnchanged(l10nent)
                        unchanged += 1
                        unchanged_w += refent.count_words()
                    else:
                        self.doChanged(ref_file, refent, l10nent)
                        changed += 1
                        changed_w += refent.count_words()
                        # run checks:
                if checker:
                    for tp, pos, msg, cat in checker.check(refent, l10nent):
                        line, col = l10nent.value_position(pos)
                        # skip error entities when merging
                        if tp == 'error' and merge_file is not None:
                            skips.append(l10nent)
                        self.notify(tp, l10n,
                                    u"%s at line %d, column %d for %s" %
                                    (msg, line, col, refent.key))
                pass

        if merge_file is not None:
            self.merge(
                ref_entities, ref_map, ref_file,
                l10n, merge_file, missings, skips, l10n_ctx,
                p.capabilities, p.encoding)

        stats = {}
        for cat, value in (
                ('missing', missing),
                ('missing_w', missing_w),
                ('report', report),
                ('obsolete', obsolete),
                ('changed', changed),
                ('changed_w', changed_w),
                ('unchanged', unchanged),
                ('unchanged_w', unchanged_w),
                ('keys', keys)):
            if value:
                stats[cat] = value
        self.updateStats(l10n, stats)
        pass

    def add(self, orig, missing):
        if self.notify('missingFile', missing, None) == "ignore":
            # filter said that we don't need this file, don't count it
            return
        f = orig
        try:
            p = parser.getParser(f.file)
        except UserWarning:
            return
        try:
            p.readContents(f.getContents())
            entities, map = p.parse()
        except Exception, ex:
            self.notify('error', f, str(ex))
            return
        # strip parse errors
        entities = [e for e in entities if not isinstance(e, parser.Junk)]
        self.updateStats(missing, {'missingInFiles': len(entities)})
        missing_w = 0
        for e in entities:
            missing_w += e.count_words()
        self.updateStats(missing, {'missing_w': missing_w})

    def doUnchanged(self, entity):
        # overload this if needed
        pass

    def doChanged(self, file, ref_entity, l10n_entity):
        # overload this if needed
        pass


def compareProjects(
            project_configs,
            stat_observer=None,
            file_stats=False,
            merge_stage=None,
            clobber_merge=False,
            quiet=0,
        ):
    locales = set()
    observers = []
    for project in project_configs:
        observers.append(
            Observer(
                quiet=quiet,
                filter=project.filter,
                file_stats=file_stats,
            ))
        locales.update(project.locales)
    if stat_observer is not None:
        stat_observers = [stat_observer]
    else:
        stat_observers = None
    comparer = ContentComparer(observers, stat_observers=stat_observers)
    for locale in sorted(locales):
        files = paths.ProjectFiles(locale, project_configs,
                                   mergebase=merge_stage)
        root = mozpath.commonprefix([m['l10n'].prefix for m in files.matchers])
        if merge_stage is not None:
            if clobber_merge:
                mergematchers = set(_m.get('merge') for _m in files.matchers)
                mergematchers.discard(None)
                for matcher in mergematchers:
                    clobberdir = matcher.prefix
                    if os.path.exists(clobberdir):
                        shutil.rmtree(clobberdir)
                        print "clobbered " + clobberdir
        for l10npath, refpath, mergepath, extra_tests in files:
            # module and file path are needed for legacy filter.py support
            module = None
            fpath = mozpath.relpath(l10npath, root)
            for _m in files.matchers:
                if _m['l10n'].match(l10npath):
                    if _m['module']:
                        # legacy ini support, set module, and resolve
                        # local path against the matcher prefix,
                        # which includes the module
                        module = _m['module']
                        fpath = mozpath.relpath(l10npath, _m['l10n'].prefix)
                    break
            reffile = paths.File(refpath, fpath or refpath, module=module)
            l10n = paths.File(l10npath, fpath or l10npath,
                              module=module, locale=locale)
            if not os.path.exists(l10npath):
                comparer.add(reffile, l10n)
                continue
            if not os.path.exists(refpath):
                comparer.remove(l10n)
                continue
            comparer.compare(reffile, l10n, mergepath, extra_tests)
    return observers
